#
# Module for starting a process object using os.fork() or CreateProcess()
#
# processing/forking.py
#
# Copyright (c) 2006-2008, R Oudkerk --- see COPYING.txt
#

import os
import sys
import signal
import processing

__all__ = ['Popen', 'assertSpawning', 'exit']

#
# Check that the current thread is spawining a child process
#

def assertSpawning(self):
    if not thisThreadIsSpawning():
        raise RuntimeError, \
              ('%s objects should only be shared between '
               'processes through inheritance' % type(self).__name__)

#
# Unix
#

if sys.platform != 'win32':
    import time
    import errno

    exit = os._exit

    def thisThreadIsSpawning():
        return False

    #
    # We define a Popen class similar to the one from subprocess, but
    # whose constructor takes a process object as its argument, and which
    # has terminate() and waitTimeout() methods.
    #

    class Popen(object):

        def __init__(self, process_obj):
            sys.stdout.flush()
            sys.stderr.flush()
            self.returncode = None

            self.pid = os.fork()
            if self.pid == 0:                
                if 'random' in sys.modules:
                    import random
                    random.seed()

                code = process_obj._bootstrap()

                sys.stdout.flush()
                sys.stderr.flush()
                os._exit(code)

        def wait(self):
            return self.poll(0)

        def poll(self, flag=os.WNOHANG):
            if self.returncode is None:
                pid, sts = os.waitpid(self.pid, flag)
                if pid == self.pid:
                    if os.WIFSIGNALED(sts):
                        self.returncode = -os.WTERMSIG(sts)
                    else:
                        assert os.WIFEXITED(sts)
                        self.returncode = os.WEXITSTATUS(sts)
            return self.returncode

        def waitTimeout(self, timeout):
            deadline = time.time() + timeout
            delay = 0.0005
            while 1:
                res = self.poll()
                if res is not None:
                    break
                remaining = deadline - time.time()
                if remaining <= 0:
                    break
                delay = min(delay * 2, remaining, 0.05)
                time.sleep(delay)
            return res

        def terminate(self):
            if self.returncode is None:
                try:
                    os.kill(self.pid, signal.SIGTERM)
                except OSError, e:
                    if self.waitTimeout(0.1) is None:
                        raise
                    
#
# Windows
#

else:
    import imp, thread, msvcrt, _subprocess
    from os.path import dirname, splitext, basename, abspath
    try:
      from dill import dump, load, HIGHEST_PROTOCOL
    except ImportError:
      from cPickle import dump, load, HIGHEST_PROTOCOL
    from processing._processing import win32
    from processing._processing import _hInterruptEvent, _main_thread_ident
    from processing.finalize import Finalize

    TERMINATE = 0x10000
    WINEXE = (sys.platform == 'win32' and getattr(sys, 'frozen', False))

    exit = win32.ExitProcess
    tls = thread._local()

    def thisThreadIsSpawning():
        return getattr(tls, 'is_spawning', False)

    #
    # We define a Popen class similar to the one from subprocess, but
    # whose constructor takes a process object as its argument, and which
    # has terminate() and waitTimeout() methods.
    #

    class Popen(object):
        '''
        Start a subprocess to run the code of a process object
        '''
        def __init__(self, process_obj):
            # create pipe for communication with child
            r, w = os.pipe()

            # get handle for read end of the pipe and make it inheritable
            rhandle = msvcrt.get_osfhandle(r)
            win32.SetHandleInformation(
                rhandle, win32.HANDLE_FLAG_INHERIT, win32.HANDLE_FLAG_INHERIT
                )

            # start process
            cmd = getCommandLine() + [rhandle]
            cmd = ' '.join('"%s"' % x for x in cmd)
            hp, ht, pid, tid = _subprocess.CreateProcess(
                sys.executable, cmd, None, None, 1, 0, None, None, None
                )
            os.close(r)
            ht.Close()

            # set attributes of self
            self.pid = pid
            self.returncode = None
            self._handle = hp

            # send information to child
            prep_data = getPreparationData(process_obj._name)
            to_child = os.fdopen(w, 'wb')
            tls.is_spawning = True
            try:
                dump(prep_data, to_child, HIGHEST_PROTOCOL)
                dump(process_obj, to_child, HIGHEST_PROTOCOL)
            finally:
                tls.is_spawning = False
                to_child.close()

        def waitTimeout(self, timeout):            
            if self.returncode is None:
                if timeout is None:
                    msecs = win32.INFINITE
                else:
                    msecs = int(timeout * 1000 + 0.5)

                if _main_thread_ident == thread.get_ident():
                    win32.ResetEvent(_hInterruptEvent)
                    handles = (int(self._handle), _hInterruptEvent)
                else:
                    handles = (int(self._handle),)

                res = win32.WaitForMultipleObjects(
                    len(handles), handles, False, msecs
                    )

                if res == win32.WAIT_OBJECT_0:
                    code = win32.GetExitCodeProcess(int(self._handle))
                    if code == TERMINATE:
                        code = -signal.SIGTERM
                    self.returncode = code

            return self.returncode

        def wait(self):
            return self.waitTimeout(None)

        def poll(self):
            return self.waitTimeout(0)

        def terminate(self):
            if self.returncode is None:
                try:
                    win32.TerminateProcess(int(self._handle), TERMINATE)
                except WindowsError:
                    if self.waitTimeout(0.1) is None:
                        raise
        
    #
    #
    #

    def isForking(argv):
        '''
        Return whether commandline indicates we are forking
        '''
        if len(argv) >= 2 and argv[1] == '--processing-fork':
            assert len(argv) == 3
            return True
        else:
            return False


    def freezeSupport():
        '''
        Run code for process object if this in not the main process
        '''
        if isForking(sys.argv):
            main()
            sys.exit()


    def getCommandLine():
        '''
        Returns prefix of command line used for spawning a child process
        '''
        if processing.currentProcess()._identity==() and isForking(sys.argv):
            raise RuntimeError, '''
            Attempt to start a new process before the current process
            has finished its bootstrapping phase.

            This probably means that you are on Windows and you have
            forgotten to use the proper idiom in the main module:

                if __name__ == '__main__':
                    freezeSupport()
                    ...

            The "freezeSupport()" line can be omitted if the program
            is not going to be frozen to produce a Windows executable.'''

        prog = 'from processing.forking import main; main()'
        if getattr(sys, 'frozen', False):
            return [sys.executable, '--processing-fork']
        elif sys.executable.lower().endswith('pythonservice.exe'):
            exe = os.path.join(os.path.dirname(os.__file__),'..','python.exe')
            return [exe, '-c', prog, '--processing-fork']
        else:
            return [sys.executable, '-c', prog, '--processing-fork']


    def getPreparationData(name):
        '''
        Return info about parent needed by child to unpickle process object
        '''
        from processing.logger import _logger
        
        if _logger is not None:
            log_args = (_logger.getEffectiveLevel(),) + _logger._extra_args
        else:
            log_args = None

        if sys.argv[0] not in ('', '-c') and not WINEXE:
            mainpath = getattr(sys.modules['__main__'], '__file__', None)
            if mainpath is not None and not os.path.isabs(mainpath):
                # we will assume os.chdir() was not used between program
                # start up and the first import of processing
                mainpath = os.path.join(processing.ORIGINAL_DIR, mainpath)
        else:
            mainpath = None
        return [name, mainpath, sys.path, sys.argv,
                processing.currentProcess().getAuthKey(),
                None, processing.ORIGINAL_DIR, log_args]


    def prepare(name, mainpath, sys_path, sys_argv, authkey,
                cur_dir, orig_dir, log_args):
        '''
        Try to get this process ready to unpickle process object
        '''
        global original_main_module

        original_main_module = sys.modules['__main__']
        processing.currentProcess().setName(name)
        processing.currentProcess().setAuthKey(authkey)

        if log_args is not None:
            from processing.logger import enableLogging
            enableLogging(*log_args)

        if orig_dir is not None:
            processing.ORIGINAL_DIR = orig_dir

        if cur_dir is not None:
            try:
                os.chdir(cur_dir)
            except OSError:
                raise

        if sys_path is not None:
            sys.path = sys_path

        if mainpath is not None:
            mainname = splitext(basename(mainpath))[0]
            if mainname == '__init__':
                mainname = basename(dirname(mainpath))

            if not mainpath.lower().endswith('.exe') and mainname != 'ipython':
                if mainpath is None:
                    dirs = None
                elif basename(mainpath).startswith('__init__.py'):
                    dirs = [dirname(dirname(mainpath))]
                else:
                    dirs = [dirname(mainpath)]

                assert mainname not in sys.modules, mainname
                file, pathname, etc = imp.find_module(mainname, dirs)
                try:
                    # We would like to do "imp.load_module('__main__', ...)"
                    # here.  However, that would cause 'if __name__ ==
                    # "__main__"' clauses to be executed.
                    main_module = imp.load_module(
                        '__parents_main__', file, pathname, etc
                        )
                finally:
                    if file:
                        file.close()

                sys.modules['__main__'] = main_module
                main_module.__name__ = '__main__'

                # XXX Try to make the potentially picklable objects in
                # sys.modules['__main__'] realize they are in the main
                # module -- ugly
                for obj in main_module.__dict__.values():
                    try:
                        if obj.__module__ == '__parents_main__':
                            obj.__module__ = '__main__'
                    except (KeyboardInterrupt, SystemExit):
                        raise
                    except:
                        pass

        if sys_argv is not None:            # this needs to come last 
            sys.argv = sys_argv

    def main():
        '''
        Run code specifed by data received over pipe
        '''
        assert isForking(sys.argv)

        handle = int(sys.argv[-1])
        fd = msvcrt.open_osfhandle(handle, os.O_RDONLY)
        from_parent = os.fdopen(fd, 'rb')

        processing.currentProcess()._inheriting = True    
        preparation_data = load(from_parent)
        prepare(*preparation_data)
        self = load(from_parent)
        processing.currentProcess()._inheriting = False

        from_parent.close()

        exitcode = self._bootstrap()
        win32.ExitProcess(exitcode)



#
# Module to support the pickling of different types of connection
# objects and file objects so that they can be transferred between
# different processes.
#
# processing/reduction.py
#
# Copyright (c) 2006-2008, R Oudkerk --- see COPYING.txt
#

__all__ = []

import os
import sys
import socket
import threading
import copy_reg
import processing

from processing import _processing
from processing.logger import debug, subDebug, subWarning
from processing.forking import thisThreadIsSpawning
from processing.process import _registerAfterFork

#
#
#

connections_are_picklable = (
    sys.platform == 'win32' or hasattr(_processing, 'recvFd')
    )

try:
    fromfd = socket.fromfd
except AttributeError:
    def fromfd(fd, family, type, proto=0):
        s = socket._socket.socket()
        _processing.changeFd(s, fd, family, type, proto)
        return s

#
# Platform specific definitions
#

if sys.platform == 'win32':
    import _subprocess
    from processing._processing import win32

    closeHandle = win32.CloseHandle
    
    def duplicateHandle(handle):
        return _subprocess.DuplicateHandle(
            _subprocess.GetCurrentProcess(), handle,
            _subprocess.GetCurrentProcess(),
            0, False, _subprocess.DUPLICATE_SAME_ACCESS
            ).Detach()

    def sendHandle(conn, handle, destination_pid):
        process_handle = win32.OpenProcess(
            win32.PROCESS_ALL_ACCESS, False, destination_pid
            )
        try:
            new_handle = _subprocess.DuplicateHandle(
                _subprocess.GetCurrentProcess(), handle, 
                process_handle, 0, False, _subprocess.DUPLICATE_SAME_ACCESS
                )
            conn.send(new_handle.Detach())
        finally:
            win32.CloseHandle(process_handle)
            
    def recvHandle(conn):
        return conn.recv()

    def isInheritableHandle(handle):
        return (win32.GetHandleInformation(handle) & win32.HANDLE_FLAG_INHERIT)

else:
    closeHandle = os.close
    duplicateHandle = os.dup
    
    def sendHandle(conn, handle, destination_pid):
        _processing.sendFd(conn.fileno(), handle)
        
    def recvHandle(conn):
        return _processing.recvFd(conn.fileno())
    
    def isInheritableHandle(handle):
        return True

#
# Support for a per-process server thread which caches pickled handles
#

_cache = set()

def _reset(obj):
    global _lock, _listener, _cache
    for h in _cache:
        closeHandle(h)
    _cache.clear()
    _lock = threading.Lock()
    _listener = None

_reset(None)
_registerAfterFork(_reset, _reset)

def _getListener():
    global _listener

    if _listener is None:
        _lock.acquire()
        try:
            if _listener is None:
                from processing.connection import Listener
                debug('starting listener and thread for sending handles')
                _listener = Listener(authenticate=True)
                t = threading.Thread(target=_serve)
                t.setDaemon(True)
                t.start()
        finally:
            _lock.release()

    return _listener

def _serve():
    while 1:
        try:
            conn = _listener.accept()
            handle_wanted, destination_pid = conn.recv()
            _cache.remove(handle_wanted)
            sendHandle(conn, handle_wanted, destination_pid)
            closeHandle(handle_wanted)
            conn.close()
        except (SystemExit, KeyboardInterrupt):
            raise
        except:
            if not processing.currentProcess()._exiting:
                import traceback
                subWarning(
                    'thread for sharing handles raised exception :\n' +
                    '-'*79 + '\n' + traceback.format_exc() + '-'*79
                    )
    
#
# Functions to be used for pickling/unpickling objects with handles
#

def reduceHandle(handle):
    if thisThreadIsSpawning() and isInheritableHandle(handle):
        return (None, handle, True)
    dup_handle = duplicateHandle(handle)
    _cache.add(dup_handle)
    subDebug('reducing handle %d', handle)
    return (_getListener().address, dup_handle, False)

def rebuildHandle(pickled_data):
    from processing.connection import Client
    address, handle, inherited = pickled_data
    if inherited:
        return handle
    subDebug('rebuilding handle %d', handle)
    conn = Client(address, authenticate=True)
    conn.send((handle, os.getpid()))
    new_handle = recvHandle(conn)
    conn.close()
    return new_handle

#
# Register `_processing.Connection` with `copy_reg`
#

def reduceConnection(conn):
    return rebuildConnection, (reduceHandle(conn.fileno()),)

def rebuildConnection(reduced_handle):
    fd = rebuildHandle(reduced_handle)
    return _processing.Connection(fd, duplicate=False)

copy_reg.pickle(_processing.Connection, reduceConnection)

#
# Register `socket.socket` with `copy_reg`
#

def reduceSocket(s):
    try:
        Family, Type, Proto = s.family, s.type, s.proto
    except AttributeError:
        # have to guess family, type, proto
        address = s.getsockname()
        Family = type(address) is str and socket.AF_UNIX or socket.AF_INET
        Type = s.getsockopt(socket.SOL_SOCKET, socket.SO_TYPE)
        Proto = 0
    reduced_handle = reduceHandle(s.fileno())
    return rebuildSocket, (reduced_handle, Family, Type, Proto)

def rebuildSocket(reduced_handle, family, type, proto):
    fd = rebuildHandle(reduced_handle)
    _sock = fromfd(fd, family, type, proto)
    closeHandle(fd)
    return socket.socket(_sock=_sock)

copy_reg.pickle(socket.socket, reduceSocket)

#
# Register `_processing.PipeConnection` with `copy_reg`
#

if sys.platform == 'win32':
    
    def reducePipeConnection(conn):
        return rebuildPipeConnection, (reduceHandle(conn.fileno()),)
    
    def rebuildPipeConnection(reduced_handle):
        handle = rebuildHandle(reduced_handle)
        return _processing.PipeConnection(handle, duplicate=False)
    
    copy_reg.pickle(_processing.PipeConnection, reducePipeConnection)



#
# Module providing the `Process` class which emulates `threading.Thread`
#
# processing/process.py
#
# Copyright (c) 2006-2008, R Oudkerk --- see COPYING.txt
#

__all__ = [
    'Process', 'currentProcess', 'activeChildren'
    ]

#
# Imports
#

import os
import sys
import time
import signal
import atexit
import weakref
import copy_reg
import itertools

#
# Public functions
#

def currentProcess():
    '''
    Return process object representing the current process
    '''
    return _current_process

def activeChildren():
    '''
    Return list of process objects corresponding to live child processes
    '''
    _cleanup()
    return list(_current_process._children)
    
#
#
#

def _cleanup():
    '''
    Purge `_children` of dead processes
    '''
    for p in list(_current_process._children):
        if p._popen.poll() is not None:
            _current_process._children.discard(p)

#
# The `Process` class
#

class Process(object):
    '''
    Process objects represent activity that is run in a separate process

    The class is analagous to `threading.Thread`
    '''
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        counter = _current_process._counter.next()

        self._identity = _current_process._identity + (counter,)
        self._authkey = _current_process._authkey
        self._daemonic = _current_process._daemonic
        self._parent_pid = os.getpid()
        self._popen = None
        self._exiting = False

        self._target = target
        self._args = tuple(args)
        self._kwargs = kwargs.copy()
        self._name = name or 'Process-' + ':'.join(map(str, self._identity))

    def run(self):
        '''
        Method to be run in sub-process; can be overridden in sub-class
        '''
        if self._target:
            self._target(*self._args, **self._kwargs)
            
    def start(self):
        '''
        Start child process
        '''
        from processing.forking import Popen
        assert self._popen is None, 'cannot start a process twice'
        assert self._parent_pid == os.getpid(), \
               'can only start a process object created by current process'
        _cleanup()
        self._popen = Popen(self)
        _current_process._children.add(self)

    def terminate(self):
        '''
        Terminate process; sends `SIGTERM` signal or uses `TerminateProcess()`
        '''
        self._popen.terminate()
        
    def join(self, timeout=None):
        '''
        Wait until child process terminates
        '''
        assert self._parent_pid == os.getpid(), 'can only join a child process'
        assert self._popen is not None, 'can only join a started process'
        if timeout == 0:
            res = self._popen.poll()
        elif timeout is None:
            res = self._popen.wait()
        else:
            res = self._popen.waitTimeout(timeout)
        if res is not None:
            _current_process._children.discard(self)

    def isAlive(self):
        '''
        Return whether child process is alive
        '''
        if self is _current_process:
            return True
        assert self._parent_pid == os.getpid(), 'can only test a child process'
        if self._popen is None:
            return False
        self._popen.poll()
        return self._popen.returncode is None

    def getName(self):
        '''
        Return name of process
        '''
        return self._name

    def setName(self, name):
        '''
        Set name of process
        '''
        assert type(name) is str, 'name must be a string'
        self._name = name

    def isDaemon(self):
        '''
        Return whether process is a daemon
        '''
        return self._daemonic

    def setDaemon(self, daemonic):
        '''
        Set whether process is a daemon
        '''
        assert self._popen is None, 'process has already started'
        self._daemonic = daemonic

    def getAuthKey(self):
        '''
        Return authorization key of process
        '''
        return self._authkey

    def setAuthKey(self, authkey):
        '''
        Set authorization key of process
        '''
        assert type(authkey) is str, 'value must be a string'
        self._authkey = authkey

    def getExitCode(self):
        '''
        Return exit code of process or `None` if it has yet to stop
        '''
        if self._popen is None:
            return self._popen
        return self._popen.poll()

    def getPid(self):
        '''
        Return PID of process or `None` if it has yet to start
        '''
        if self is _current_process:
            return os.getpid()
        else:
            assert self._parent_pid == os.getpid(), 'not a child process'
            return self._popen and self._popen.pid

    def __repr__(self):
        if self is _current_process:
            status = 'started'
        elif self._parent_pid != os.getpid():
            status = 'unknown'
        elif self._popen is None:
            status = 'initial'
        else:
            if self._popen.poll() is not None:
                status = self.getExitCode()
            else:
                status = 'started'

        if type(status) is int:
            if status == 0:
                status = 'stopped'
            else:
                status = 'stopped[%s]' % _exitcode_to_name.get(status, status)

        return '<%s(%s, %s%s)>' % (type(self).__name__, self._name,
                                   status, self._daemonic and ' daemon' or '')

    ##

    def _bootstrap(self):
        from processing.finalize import _registry
        from processing.logger import info
        
        global _current_process
        try:
            self._children = set()
            self._counter = itertools.count(1)
            sys.stdin.close()
            _registry.clear()
            _current_process = self
            _runAfterForkers()
            info('child process calling self.run()')
            try:
                self.run()
                exitcode = 0
            finally:
                _exitFunction()
        except SystemExit, e:
            if not e.args:
                exitcode = 1
            elif type(e.args[0]) is int:
                exitcode = e.args[0]
            else:
                print >>sys.stderr, e.args[0]
                exitcode = 1
        except:
            exitcode = 1
            import traceback
            print >>sys.stderr, 'Process %s:' % self.getName()
            traceback.print_exc()

        info('process exiting with exitcode %d' % exitcode)
        return exitcode

#
# Create object representing the main process
#

class _MainProcess(Process):

    def __init__(self):
        self._identity = ()
        self._daemonic = False
        self._name = 'MainProcess'
        self._parent_pid = None
        self._popen = None
        self._counter = itertools.count(1)
        self._children = set()
        self._authkey = ''.join('%02x' % ord(c) for c in os.urandom(16))


_current_process = _MainProcess()
del _MainProcess

#
# Give names to some return codes
#

_exitcode_to_name = {}

for name, signum in signal.__dict__.items():
    if name[:3]=='SIG' and '_' not in name:
        _exitcode_to_name[-signum] = name

#
# Make bound and unbound instance methods and class methods picklable
#

def _reduceMethod(m):
    if m.im_self is None:
        return getattr, (m.im_class, m.im_func.func_name)
    else:
        return getattr, (m.im_self, m.im_func.func_name)

copy_reg.pickle(type(_current_process.start), _reduceMethod)

#
# Support for reinitialization of objects when bootstrapping a child process
#

_afterfork_registry = weakref.WeakValueDictionary()
_afterForkerId = itertools.count().next

def _runAfterForkers():
    # execute in order of registration
    for (index, ident, func), obj in sorted(_afterfork_registry.items()):
        func(obj)
        
def _registerAfterFork(obj, func):
    _afterfork_registry[(_afterForkerId(), id(obj), func)] = obj
    
#
# Clean up on exit
#

def _exitFunction():
    from processing.finalize import _runFinalizers
    from processing.logger import info
    
    _current_process._exiting = True

    info('running all "atexit" finalizers with priority >= 0')
    _runFinalizers(0)

    for p in activeChildren():
        if p._daemonic:
            info('calling `terminate()` for daemon %s', p.getName())
            p._popen.terminate()

    for p in activeChildren():
        info('calling `join()` for process %s', p.getName())
        p.join()

    info('running the remaining "atexit" finalizers')
    _runFinalizers()

atexit.register(_exitFunction)



#
# Module implementing synchronization primitives
#
# processing/synchronize.py
#
# Copyright (c) 2006-2008, R Oudkerk --- see COPYING.txt
#

__all__ = [ 'Lock', 'RLock', 'Semaphore', 'BoundedSemaphore', 'Condition',
            'Event']

import threading
import os
import sys

from struct import pack as _pack, unpack as _unpack, calcsize as _calcsize
from time import time as _time, sleep as _sleep

from processing import _processing
from processing.process import currentProcess, _registerAfterFork
from processing.logger import debug
from processing.finalize import Finalize
from processing.forking import assertSpawning

#
# Constants to describe the kind of blocker (normal lock is a bounded sem)
#

RECURSIVE_MUTEX, SEMAPHORE, BOUNDED_SEMAPHORE = range(3)

#
# Base class for semaphores and mutexes; wraps `_processing.SemLock`
#

class SemLock(object):

    def __init__(self, kind, value):
        sl = self._semlock = _processing.SemLock(kind, value)
        debug('created semlock with handle %s' % sl.handle)
        self.__setstate__((sl.handle, sl.kind, sl.maxvalue))

        if sys.platform != 'win32':
            def _afterFork(obj):
                obj._semlock._afterFork()
            _registerAfterFork(self, _afterFork)

    def __getstate__(self):
        assertSpawning(self)
        return self._state

    def __setstate__(self, state):
        self._state = state
        if not hasattr(self, '_semlock'):
            self._semlock = _processing.SemLock._rebuild(*state)
            debug('recreated blocker with handle %r' % state[0])

        self.acquire = self._semlock.acquire
        self.release = self._semlock.release
        self.__enter__ = self._semlock.__enter__
        self.__exit__ = self._semlock.__exit__

#
# Semaphore
#

class Semaphore(SemLock):

    def __init__(self, value=1):
        SemLock.__init__(self, SEMAPHORE, value)

    def getValue(self):
        return self._semlock._getValue()

    def __repr__(self):
        try:
            value = self._semlock._getValue()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            value = 'unknown'
        return '<Semaphore(value=%s)>' % value

#
# Bounded semaphore
#

class BoundedSemaphore(Semaphore):

    def __init__(self, value=1):
        SemLock.__init__(self, BOUNDED_SEMAPHORE, value)

    def __repr__(self):
        try:
            value = self._semlock._getValue()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            value = 'unknown'
        return '<BoundedSemaphore(value=%s, maxvalue=%s)>' % \
               (value, self._semlock.maxvalue)

#
# Non-recursive lock
#

class Lock(SemLock):

    def __init__(self):
        SemLock.__init__(self, BOUNDED_SEMAPHORE, 1)

    def __repr__(self):
        try:
            if self._semlock._isMine():
                name = currentProcess().getName()
                if threading.currentThread().getName() != 'MainThread':
                    name += '|' + threading.currentThread().getName()
            elif self._semlock._getValue() == 1:
                name = 'None'
            elif self._semlock._count() > 0:
                name = 'SomeOtherThread'
            else:
                name = 'SomeOtherProcess'
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            name = 'unknown'
        return '<Lock(owner=%s)>' % name

#
# Recursive lock
#

class RLock(SemLock):

    def __init__(self):
        SemLock.__init__(self, RECURSIVE_MUTEX, 1)

    def __repr__(self):
        try:
            if self._semlock._isMine():
                name = currentProcess().getName()
                if threading.currentThread().getName() != 'MainThread':
                    name += '|' + threading.currentThread().getName()
                count = self._semlock._count()
            elif self._semlock._getValue() == 1:
                name, count = 'None', 0
            elif self._semlock._count() > 0:
                name, count = 'SomeOtherThread', 'nonzero'
            else:
                name, count = 'SomeOtherProcess', 'nonzero'
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            name, count = 'unknown', 'unknown'
        return '<RLock(%s, %s)>' % (name, count)

#
# Condition variable
#

class Condition(object):

    def __init__(self, lock=None):
        state = (lock or RLock(), Semaphore(0), Semaphore(0), Semaphore(0))
        self.__setstate__(state)

    def __getstate__(self):
        assertSpawning(self)
        return self._state

    def __setstate__(self, state):
        (self._lock, self._sleeping_count,
         self._woken_count, self._wait_semaphore) = self._state = state
        self.acquire = self._lock.acquire
        self.release = self._lock.release
        self.__enter__ = self._lock.__enter__
        self.__exit__ = self._lock.__exit__

    def __repr__(self):
        try:
            num_waiters = (self._sleeping_count._semlock._getValue() -
                           self._woken_count._semlock._getValue())
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            num_waiters = 'unkown'
        return '<Condition(%s, %s)>' % (self._lock, num_waiters)

    def wait(self, timeout=None):
        assert self._lock._semlock._isMine(), \
               'must acquire() condition before using wait()'

        # indicate that this thread is going to sleep
        self._sleeping_count.release()

        # release lock
        count = self._lock._semlock._count()
        for i in xrange(count):
            self._lock.release()

        try:
            # wait for notification or timeout
            self._wait_semaphore.acquire(True, timeout)
        finally:
            # indicate that this thread has woken
            self._woken_count.release()

            # reacquire lock
            for i in xrange(count):
                self._lock.acquire()

    def notify(self):
        assert self._lock._semlock._isMine(), 'lock is not owned'
        assert not self._wait_semaphore.acquire(False)
        
        # to take account of timeouts since last notify() we subtract
        # woken_count from sleeping_count and rezero woken_count
        while self._woken_count.acquire(False):
            res = self._sleeping_count.acquire(False)
            assert res
            
        if self._sleeping_count.acquire(False): # try grabbing a sleeper
            self._wait_semaphore.release()      # wake up one sleeper
            self._woken_count.acquire()         # wait for the sleeper to wake
            
            # rezero _wait_semaphore in case a timeout just happened
            self._wait_semaphore.acquire(False)

    def notifyAll(self):
        assert self._lock._semlock._isMine(), 'lock is not owned'
        assert not self._wait_semaphore.acquire(False)

        # to take account of timeouts since last notify*() we subtract
        # woken_count from sleeping_count and rezero woken_count
        while self._woken_count.acquire(False):
            res = self._sleeping_count.acquire(False)
            assert res
            
        sleepers = 0
        while self._sleeping_count.acquire(False):
            self._wait_semaphore.release()        # wake up one sleeper
            sleepers += 1

        if sleepers:
            for i in xrange(sleepers):
                self._woken_count.acquire()       # wait for a sleeper to wake

            # rezero wait_semaphore in case some timeouts just happened
            while self._wait_semaphore.acquire(False):
                pass

#
# Event
#

class Event(object):

    def __init__(self):
        self._cond = Condition(Lock())
        self._flag = Semaphore(0)

    def isSet(self):
        self._cond.acquire()
        try:
            if self._flag.acquire(False):
                self._flag.release()
                return True
            return False
        finally:
            self._cond.release()
    
    def set(self):
        self._cond.acquire()
        try:
            self._flag.acquire(False)
            self._flag.release()
            self._cond.notifyAll()
        finally:
            self._cond.release()

    def clear(self):
        self._cond.acquire()
        try:
            self._flag.acquire(False)
        finally:
            self._cond.release()

    def wait(self, timeout=None):
        self._cond.acquire()
        try:
            if self._flag.acquire(False):
                self._flag.release()
            else:
                self._cond.wait(timeout)
        finally:
            self._cond.release()



#
# Module supporting finaliztion using weakrefs
#
# processing/finalize.py
#
# Copyright (c) 2006-2008, R Oudkerk --- see COPYING.txt
#

import weakref
import itertools

from processing.logger import subDebug

__all__ = ['Finalize', '_runFinalizers']


_registry = {}
_counter = itertools.count()


class Finalize(object):
    '''
    Class which supports object finalization using weakrefs
    '''  
    def __init__(self, obj, callback, args=(), kwargs=None, exitpriority=None):
        assert exitpriority is None or type(exitpriority) is int
        assert callback is not None
        
        if obj is not None:
            self._weakref = weakref.ref(obj, self)
        else:
            assert exitpriority is not None

        self._callback = callback
        self._args = args
        self._kwargs = kwargs or {}
        self._key = (exitpriority, _counter.next())
        
        _registry[self._key] = self
        
    def __call__(self, wr=None):
        '''
        Run the callback unless it has already been called or cancelled

        Returns True if callback was run otherwise returns False
        '''
        try:
            del _registry[self._key]
        except KeyError:
            subDebug('finalizer no longer registered')
        else:
            subDebug('finalizer calling %s with args %s and kwargs %s',
                     self._callback, self._args, self._kwargs)
            self._callback(*self._args, **self._kwargs)
            self._weakref = self._callback = self._args = \
                            self._kwargs = self._key = None
            return True

    def cancel(self):
        '''
        Cancel finalization of the object
        '''
        try:
            del _registry[self._key]
        except KeyError:
            pass
        else:
            self._weakref = self._callback = self._args = \
                            self._kwargs = self._key = None

    def stillActive(self):
        '''
        Return whether this finalizer is still waiting to invoke callback
        '''
        return self._key in _registry

    def __repr__(self):
        try:
            obj = self._weakref()
        except (AttributeError, TypeError):
            obj = None

        if obj is None:
            return '<Finalize object, dead>'

        x = '<Finalize object, callback=%s' % \
            getattr(self._callback, '__name__', self._callback)
        if self._args:
            x += ', args=' + str(self._args)
        if self._kwargs:
            x += ', kwargs=' + str(self._kwargs)
        if self._key[0] is not None:
            x += ', exitprority=' + str(self._key[0])
        return x + '>'


def _runFinalizers(minpriority=None):
    '''
    Run all finalizers whose exit priority is not None and at least minpriority
    
    Finalizers with highest priority are called first; finalizers with
    the same priority will be called in reverse order of creation.
    '''
    if minpriority is None:
        f = lambda p : p[0][0] is not None
    else:
        f = lambda p : p[0][0] is not None and p[0][0] >= minpriority

    items = sorted(filter(f, _registry.items()), reverse=True)

    for key, finalizer in items:
        subDebug('calling %s', finalizer)
        try:
            finalizer()
        except Exception:
            import traceback
            traceback.print_exc()

    if minpriority is None:
        _registry.clear()



#
# Unit tests for processing package
#

import unittest
import threading
import Queue
import time
import sys
import os
import signal
import array
import copy
import socket
import random
import logging
import ctypes

import processing.dummy
import processing.connection
import processing.managers
import processing.heap
import processing.managers
import processing.pool

#
# Constants
#

DELTA = 0.1
CHECK_TIMINGS = False     # making true makes tests take a lot longer
                          # and can sometimes cause some non-serious
                          # failures because some calls block a bit
                          # longer than expected
if CHECK_TIMINGS:
    TIMEOUT1, TIMEOUT2, TIMEOUT3 = 0.82, 0.35, 1.4
else:
    TIMEOUT1, TIMEOUT2, TIMEOUT3 = 0.1, 0.1, 0.1

HAVE_GETVALUE = not getattr(processing._processing,
                            'HAVE_BROKEN_SEM_GETVALUE', False)

#
# Creates a wrapper for a function which records the time it takes to finish
#

class TimingWrapper(object):

    def __init__(self, func):
        self.func = func
        self.elapsed = None

    def __call__(self, *args, **kwds):
        t = time.time()
        try:
            return self.func(*args, **kwds)
        finally:
            self.elapsed = time.time() - t
        
#
# Base class for test cases
#

class BaseTestCase(object):
    
    ALLOWED_TYPES = ('processes', 'manager', 'threads')

    def assertTimingAlmostEqual(self, a, b):
        if CHECK_TIMINGS:
            self.assertAlmostEqual(a, b, 1)

    def assertReturnsIfImplemented(self, value, func, *args):
        try:
            res = func(*args)
        except NotImplementedError:
            pass
        else:
            return self.assertEqual(value, res)

#
# Return the value of a semaphore
#

def getValue(self):
    try:
        return self.getValue()
    except AttributeError:
        try:
            return self._Semaphore__value
        except AttributeError:
            try:
                return self._value
            except AttributeError:
                raise NotImplementedError

#
# Testcases
#

class _TestProcess(BaseTestCase):

    ALLOWED_TYPES = ('processes', 'threads')

    def test_current(self):
        current = self.currentProcess()
        self.assertTrue(current.isAlive())
        self.assertTrue(not current.isDaemon())
        if self.TYPE != 'threads':
            authkey = current.getAuthKey()
            self.assertTrue(type(authkey) is str)
            self.assertTrue(len(authkey) > 0)
            self.assertEqual(current.getPid(), os.getpid())
            self.assertEqual(current.getExitCode(), None)

    def _test(self, q, *args, **kwds):
        current = self.currentProcess()
        q.put(args)
        q.put(kwds)
        q.put(current.getName())
        if self.TYPE != 'threads':
            q.put(current.getAuthKey())
            q.put(current.getPid())

    def test_process(self):
        q = self.Queue(1)
        e = self.Event()
        args = (q, 1, 2)
        kwargs = {'hello':23, 'bye':2.54}
        name = 'SomeProcess'
        p = self.Process(
            target=self._test, args=args, kwargs=kwargs, name=name
            )
        p.setDaemon(True)
        current = self.currentProcess()

        if self.TYPE != 'threads':
            self.assertEquals(p.getAuthKey(), current.getAuthKey())
        self.assertEquals(p.isAlive(), False)
        self.assertEquals(p.isDaemon(), True)
        self.assertTrue(p not in self.activeChildren())
        self.assertTrue(type(self.activeChildren()) is list)
        self.assertEqual(p.getExitCode(), None)
        
        p.start()
        
        self.assertEquals(p.getExitCode(), None)
        self.assertEquals(p.isAlive(), True)
        self.assertTrue(p in self.activeChildren())
        
        self.assertEquals(q.get(), args[1:])
        self.assertEquals(q.get(), kwargs)
        self.assertEquals(q.get(), p.getName())
        if self.TYPE != 'threads':
            self.assertEquals(q.get(), current.getAuthKey())
            self.assertEquals(q.get(), p.getPid())

        p.join()

        self.assertEquals(p.getExitCode(), 0)
        self.assertEquals(p.isAlive(), False)
        self.assertTrue(p not in self.activeChildren())        

    def _test_terminate(self):
        time.sleep(1000)

    def test_terminate(self):
        if self.TYPE == 'threads':
            return
        
        p = self.Process(target=self._test_terminate)
        p.setDaemon(True)
        p.start()

        self.assertEqual(p.isAlive(), True)
        self.assertTrue(p in self.activeChildren())
        self.assertEqual(p.getExitCode(), None)

        p.terminate()

        join = TimingWrapper(p.join)
        self.assertEqual(join(), None)
        self.assertTimingAlmostEqual(join.elapsed, 0.0)
        
        self.assertEqual(p.isAlive(), False)
        self.assertTrue(p not in self.activeChildren())

        p.join()

        # XXX sometimes get p.getExitCode() == 0 on Windows ...
        #self.assertEqual(p.getExitCode(), -signal.SIGTERM)

    def test_cpuCount(self):
        try:
            cpus = processing.cpuCount()
        except NotImplementedError:
            cpus = 1
        self.assertTrue(type(cpus) is int)
        self.assertTrue(cpus >= 1)

    def test_activeChildren(self):
        self.assertEqual(type(self.activeChildren()), list)

        p = self.Process(target=time.sleep, args=(DELTA,))
        self.assertTrue(p not in self.activeChildren())
        
        p.start()
        self.assertTrue(p in self.activeChildren())

        p.join()
        self.assertTrue(p not in self.activeChildren())

    def _test_recursion(self, wconn, id):
        wconn.send(id)
        if len(id) < 2:
            for i in range(2):
                p = self.Process(
                    target=self._test_recursion, args=(wconn, id+[i])
                    )
                p.start()
                p.join()

    def test_recursion(self):
        rconn, wconn = self.Pipe(duplex=False)
        self._test_recursion(wconn, [])
        
        time.sleep(DELTA)
        result = []
        while rconn.poll():
            result.append(rconn.recv())
            
        expected = [
            [],
              [0],
                [0, 0],
                [0, 1],
              [1],
                [1, 0],
                [1, 1]
            ]
        self.assertEqual(result, expected)

#
#
#

class _UpperCaser(processing.Process):

    def __init__(self):
        processing.Process.__init__(self)
        self.child_conn, self.parent_conn = processing.Pipe()

    def run(self):
        self.parent_conn.close()
        for s in iter(self.child_conn.recv, None):
            self.child_conn.send(s.upper())
        self.child_conn.close()

    def submit(self, s):
        assert type(s) is str
        self.parent_conn.send(s)
        return self.parent_conn.recv()

    def stop(self):
        self.parent_conn.send(None)
        self.parent_conn.close()
        self.child_conn.close()

class _TestSubclassingProcess(BaseTestCase):

    ALLOWED_TYPES = ('processes',)

    def test_subclassing(self):
        uppercaser = _UpperCaser()
        uppercaser.start()
        self.assertEqual(uppercaser.submit('hello'), 'HELLO')
        self.assertEqual(uppercaser.submit('world'), 'WORLD')
        uppercaser.stop()
        uppercaser.join()
        
#
#
#

class _TestQueue(BaseTestCase):

    def _test_put(self, queue, child_can_start, parent_can_continue):
        child_can_start.wait()
        for i in range(6):
            queue.get()
        parent_can_continue.set()

    def test_put(self):
        queue = self.Queue(maxsize=6)
        child_can_start = self.Event()
        parent_can_continue = self.Event()

        proc = self.Process(
            target=self._test_put,
            args=(queue, child_can_start, parent_can_continue)
            )
        proc.setDaemon(True)
        proc.start()
        
        self.assertEqual(queue.empty(), True)
        self.assertEqual(queue.full(), False)

        queue.put(1)
        queue.put(2, True)
        queue.put(3, True, None)
        queue.put(4, False)
        queue.put(5, False, None)
        queue.put_nowait(6)

        # the values may be in buffer but not yet in pipe so sleep a bit
        time.sleep(DELTA)     

        self.assertEqual(queue.empty(), False)
        self.assertEqual(queue.full(), True)

        put = TimingWrapper(queue.put)
        put_nowait = TimingWrapper(queue.put_nowait)

        self.assertRaises(Queue.Full, put, 7, False)
        self.assertTimingAlmostEqual(put.elapsed, 0)

        self.assertRaises(Queue.Full, put, 7, False, None)
        self.assertTimingAlmostEqual(put.elapsed, 0)

        self.assertRaises(Queue.Full, put_nowait, 7)
        self.assertTimingAlmostEqual(put_nowait.elapsed, 0)

        self.assertRaises(Queue.Full, put, 7, True, TIMEOUT1)
        self.assertTimingAlmostEqual(put.elapsed, TIMEOUT1)

        self.assertRaises(Queue.Full, put, 7, False, TIMEOUT2)
        self.assertTimingAlmostEqual(put.elapsed, 0)

        self.assertRaises(Queue.Full, put, 7, True, timeout=TIMEOUT3)
        self.assertTimingAlmostEqual(put.elapsed, TIMEOUT3)

        child_can_start.set()
        parent_can_continue.wait()

        self.assertEqual(queue.empty(), True)
        self.assertEqual(queue.full(), False)

        proc.join()

    def _test_get(self, queue, child_can_start, parent_can_continue):
        child_can_start.wait()
        queue.put(1)
        queue.put(2)
        if self.TYPE == 'processes':
            queue.putmany([3, 4, 5])
        else:
            queue.put(3)
            queue.put(4)
            queue.put(5)
        parent_can_continue.set()
        
    def test_get(self):
        queue = self.Queue()
        child_can_start = self.Event()
        parent_can_continue = self.Event()
        
        proc = self.Process(
            target=self._test_get,
            args=(queue, child_can_start, parent_can_continue)
            )
        proc.setDaemon(True)
        proc.start()
        
        self.assertEqual(queue.empty(), True)
        
        child_can_start.set()
        parent_can_continue.wait()

        time.sleep(DELTA)
        self.assertEqual(queue.empty(), False)

        self.assertEqual(queue.get(), 1)
        self.assertEqual(queue.get(True, None), 2)
        self.assertEqual(queue.get(True), 3)
        self.assertEqual(queue.get(timeout=1), 4)
        self.assertEqual(queue.get_nowait(), 5)
        
        self.assertEqual(queue.empty(), True)

        get = TimingWrapper(queue.get)
        get_nowait = TimingWrapper(queue.get_nowait)
        
        self.assertRaises(Queue.Empty, get, False)
        self.assertTimingAlmostEqual(get.elapsed, 0)

        self.assertRaises(Queue.Empty, get, False, None)
        self.assertTimingAlmostEqual(get.elapsed, 0)

        self.assertRaises(Queue.Empty, get_nowait)
        self.assertTimingAlmostEqual(get_nowait.elapsed, 0)

        self.assertRaises(Queue.Empty, get, True, TIMEOUT1)
        self.assertTimingAlmostEqual(get.elapsed, TIMEOUT1)

        self.assertRaises(Queue.Empty, get, False, TIMEOUT2)
        self.assertTimingAlmostEqual(get.elapsed, 0)

        self.assertRaises(Queue.Empty, get, timeout=TIMEOUT3)
        self.assertTimingAlmostEqual(get.elapsed, TIMEOUT3)

        proc.join()
        
    def _test_fork(self, queue):
        for i in range(10, 20):
            queue.put(i)
        # note that at this point the items may only be buffered, so the
        # process cannot shutdown until the feeder thread has finished
        # pushing items onto the pipe.

    def test_fork(self):
        # Old versions of Queue would fail to create a new feeder
        # thread for a forked process if the original process had its
        # own feeder thread.  This test checks that this no longer
        # happens.

        queue = self.Queue()

        # put items on queue so that main process starts a feeder thread
        for i in range(10):
            queue.put(i)

        # wait to make sure thread starts before we fork a new process
        time.sleep(DELTA)

        # fork process
        p = self.Process(target=self._test_fork, args=(queue,))
        p.start()

        # check that all expected items are in the queue
        for i in range(20):
            self.assertEqual(queue.get(), i)
        self.assertRaises(Queue.Empty, queue.get, False)

        p.join()

#
#
#

class _TestLock(BaseTestCase):

    def test_lock(self):
        lock = self.Lock()
        self.assertEqual(lock.acquire(), True)
        self.assertEqual(lock.acquire(False), False)
        self.assertEqual(lock.release(), None)
        self.assertRaises((ValueError, threading.ThreadError), lock.release)

    def test_rlock(self):
        lock = self.RLock()
        self.assertEqual(lock.acquire(), True)
        self.assertEqual(lock.acquire(), True)
        self.assertEqual(lock.acquire(), True)
        self.assertEqual(lock.release(), None)
        self.assertEqual(lock.release(), None)
        self.assertEqual(lock.release(), None)
        self.assertRaises((AssertionError, RuntimeError), lock.release)
        
        
class _TestSemaphore(BaseTestCase):

    def _test_semaphore(self, sem):
        self.assertReturnsIfImplemented(2, getValue, sem)
        self.assertEqual(sem.acquire(), True)
        self.assertReturnsIfImplemented(1, getValue, sem)
        self.assertEqual(sem.acquire(), True)
        self.assertReturnsIfImplemented(0, getValue, sem)
        self.assertEqual(sem.acquire(False), False)
        self.assertReturnsIfImplemented(0, getValue, sem)
        self.assertEqual(sem.release(), None)
        self.assertReturnsIfImplemented(1, getValue, sem)
        self.assertEqual(sem.release(), None)
        self.assertReturnsIfImplemented(2, getValue, sem)
        
    def test_semaphore(self):
        sem = self.Semaphore(2)
        self._test_semaphore(sem)
        self.assertEqual(sem.release(), None)
        self.assertReturnsIfImplemented(3, getValue, sem)
        self.assertEqual(sem.release(), None)
        self.assertReturnsIfImplemented(4, getValue, sem)

    def test_bounded_semaphore(self):
        sem = self.BoundedSemaphore(2)
        self._test_semaphore(sem)
        if HAVE_GETVALUE:
            self.assertRaises(ValueError, sem.release)
            self.assertReturnsIfImplemented(2, getValue, sem)

    def test_timeout(self):
        if self.TYPE != 'processes':
            return

        sem = self.Semaphore(0)
        acquire = TimingWrapper(sem.acquire)

        self.assertEqual(acquire(False), False)
        self.assertTimingAlmostEqual(acquire.elapsed, 0.0)

        self.assertEqual(acquire(False, None), False)
        self.assertTimingAlmostEqual(acquire.elapsed, 0.0)

        self.assertEqual(acquire(False, TIMEOUT1), False)
        self.assertTimingAlmostEqual(acquire.elapsed, 0)

        self.assertEqual(acquire(True, TIMEOUT2), False)
        self.assertTimingAlmostEqual(acquire.elapsed, TIMEOUT2)

        self.assertEqual(acquire(timeout=TIMEOUT3), False)
        self.assertTimingAlmostEqual(acquire.elapsed, TIMEOUT3)


class _TestCondition(BaseTestCase):
    
    def f(self, cond, sleeping, woken, timeout=None):
        cond.acquire()
        sleeping.release()
        cond.wait(timeout)
        woken.release()
        cond.release()
    
    def check_invariant(self, cond):
        # this is only supposed to succeed when there are no sleepers
        if self.TYPE == 'processes':
            try:
                sleepers = (cond._sleeping_count.getValue() -
                            cond._woken_count.getValue())
                self.assertEqual(sleepers, 0)
                self.assertEqual(cond._wait_semaphore.getValue(), 0)
            except NotImplementedError:
                pass
            
    def test_notify(self):
        cond = self.Condition()
        sleeping = self.Semaphore(0)
        woken = self.Semaphore(0)
        
        p = self.Process(target=self.f, args=(cond, sleeping, woken))
        p.setDaemon(True)
        p.start()

        p = threading.Thread(target=self.f, args=(cond, sleeping, woken))
        p.setDaemon(True)
        p.start()
        
        # wait for both children to start sleeping
        sleeping.acquire()
        sleeping.acquire()
        
        # check no process/thread has woken up
        time.sleep(DELTA)
        self.assertReturnsIfImplemented(0, getValue, woken)

        # wake up one process/thread
        cond.acquire()
        cond.notify()
        cond.release()
        
        # check one process/thread has woken up
        time.sleep(DELTA)
        self.assertReturnsIfImplemented(1, getValue, woken)

        # wake up another
        cond.acquire()
        cond.notify()
        cond.release()
        
        # check other has woken up
        time.sleep(DELTA)
        self.assertReturnsIfImplemented(2, getValue, woken)
        
        # check state is not mucked up
        self.check_invariant(cond)
        p.join()
        
    def test_notifyAll(self):
        cond = self.Condition()
        sleeping = self.Semaphore(0)
        woken = self.Semaphore(0)

        # start some threads/processes which will timeout
        for i in range(3):
            p = self.Process(target=self.f,
                             args=(cond, sleeping, woken, TIMEOUT1))
            p.setDaemon(True)
            p.start()

            t = threading.Thread(target=self.f,
                                 args=(cond, sleeping, woken, TIMEOUT1))
            t.setDaemon(True)
            t.start()

        # wait for them all to sleep
        for i in xrange(6):
            sleeping.acquire()

        # check they have all timed out
        for i in xrange(6):
            woken.acquire()
        self.assertReturnsIfImplemented(0, getValue, woken)

        # check state is not mucked up
        self.check_invariant(cond)

        # start some more threads/processes
        for i in range(3):
            p = self.Process(target=self.f, args=(cond, sleeping, woken))
            p.setDaemon(True)
            p.start()
            
            t = threading.Thread(target=self.f, args=(cond, sleeping, woken))
            t.setDaemon(True)
            t.start()
            
        # wait for them to all sleep
        for i in xrange(6):
            sleeping.acquire()
            
        # check no process/thread has woken up
        time.sleep(DELTA)
        self.assertReturnsIfImplemented(0, getValue, woken)

        # wake them all up
        cond.acquire()
        cond.notifyAll()
        cond.release()

        # check they have all woken
        time.sleep(DELTA)
        self.assertReturnsIfImplemented(6, getValue, woken)

        # check state is not mucked up
        self.check_invariant(cond)

    def test_timeout(self):
        cond = self.Condition()
        wait = TimingWrapper(cond.wait)
        cond.acquire()
        res = wait(TIMEOUT1)
        cond.release()
        self.assertEqual(res, None)
        self.assertTimingAlmostEqual(wait.elapsed, TIMEOUT1)

        
class _TestEvent(BaseTestCase):

    def _test_event(self, event):
        time.sleep(TIMEOUT2)
        event.set()

    def test_event(self):
        event = self.Event()
        wait = TimingWrapper(event.wait)

        self.assertEqual(event.isSet(), False)
        self.assertEqual(wait(0.0), None)
        self.assertTimingAlmostEqual(wait.elapsed, 0.0)
        self.assertEqual(wait(TIMEOUT1), None)
        self.assertTimingAlmostEqual(wait.elapsed, TIMEOUT1)

        event.set()

        self.assertEqual(event.isSet(), True)
        self.assertEqual(wait(), None)
        self.assertTimingAlmostEqual(wait.elapsed, 0.0)
        self.assertEqual(wait(TIMEOUT1), None)
        self.assertTimingAlmostEqual(wait.elapsed, 0.0)
        self.assertEqual(event.isSet(), True)

        event.clear()

        self.assertEqual(event.isSet(), False)

        self.Process(target=self._test_event, args=(event,)).start()
        self.assertEqual(wait(), None)

#
#
#

class _TestValue(BaseTestCase):

    codes_values = [
        ('i', 4343, 24234),
        ('d', 3.625, -4.25),
        ('h', -232, 234),
        ('c', 'x', 'y'),
        ]

    def _test(self, values):
        for sv, cv in zip(values, self.codes_values):
            sv.value = cv[2]
        
    def test_sharedvalue(self, lock=False):
        values = [self.Value(code, value)
                  for code, value, _ in self.codes_values]

        for sv, cv in zip(values, self.codes_values):
            self.assertEqual(sv.value, cv[1])
        
        proc = self.Process(target=self._test, args=(values,))
        proc.start()
        proc.join()

        for sv, cv in zip(values, self.codes_values):
            self.assertEqual(sv.value, cv[2])

    def test_synchronized(self):
        self.test_sharedvalue(lock=True)

    def test_getobj_getlock(self):
        if self.TYPE != 'processes':
            return
        lock = self.Lock()
        obj = self.Value('i', 5, lock=lock)
        self.assertEqual(obj.getlock(), lock)
        self.assertEqual(obj.getobj().value, 5)


class _TestArray(BaseTestCase):

    def f(self, seq):
        for i in range(1, len(seq)):
            seq[i] += seq[i-1]

    def test_sharedarray(self, lock=False):
        seq = [680, 626, 934, 821, 150, 233, 548, 982, 714, 831]
        arr = self.Array('i', seq, lock=lock)
        
        self.assertEqual(len(arr), len(seq))
        self.assertEqual(arr[3], seq[3])
        self.assertEqual(list(arr[2:7]), list(seq[2:7]))
        
        arr[4:8] = seq[4:8] = array.array('i', [1, 2, 3, 4])
        
        self.assertEqual(list(arr), seq)
        
        self.f(seq)
        
        p = self.Process(target=self.f, args=(arr,))
        p.start()
        p.join()
        
        self.assertEqual(list(arr), list(seq))
        
    def test_synchronized(self):
        self.test_sharedarray(lock=True)
        
#
#
#

class _TestContainers(BaseTestCase):

    ALLOWED_TYPES = ('manager',)

    def test_list(self):
        a = self.list(range(10))
        self.assertEqual(a[:], range(10))
        
        b = self.list()
        self.assertEqual(b[:], [])
        
        b.extend(range(5))
        self.assertEqual(b[:], range(5))
        
        self.assertEqual(b[2], 2)
        self.assertEqual(b[2:10], [2,3,4])

        b *= 2
        self.assertEqual(b[:], [0, 1, 2, 3, 4, 0, 1, 2, 3, 4])

        self.assertEqual(b + [5, 6], [0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 5, 6])

        self.assertEqual(a, self.list(range(10)))
        self.assertEqual(a, range(10))
        self.assertTrue(a != range(11))
        self.assertTrue(range(9) < a)
        self.assertTrue(a < range(11))

        d = [a, b]
        e = self.list(d)
        self.assertEqual(
            e, [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [0, 1, 2, 3, 4, 0, 1, 2, 3, 4]]
            )

        it = iter(a)
        self.assertEqual(tuple(it), (0, 1, 2, 3, 4, 5, 6, 7, 8, 9))
        
        f = self.list([a])
        a.append('hello')
        self.assertEqual(f, [[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 'hello']])

    def test_dict(self):
        d = self.dict()
        for i in range(5):
            d[i] = chr(65 + i)
        self.assertEqual(
            d, {0: 'A', 1: 'B', 2: 'C', 3: 'D', 4: 'E'}
            )
        self.assertEqual(list(d), range(5))
        self.assertEqual(
            list(d.iteritems()),
            [(0, 'A'), (1, 'B'), (2, 'C'), (3, 'D'), (4, 'E')]
            )
        
    def test_namespace(self):
        n = self.Namespace()
        n.name = 'Bob'
        n.job = 'Builder'
        self.assertEqual((n.name, n.job), ('Bob', 'Builder'))
        del n.job
        self.assertEqual(str(n), "Namespace(name='Bob')")
        self.assertTrue(hasattr(n, 'name'))
        self.assertTrue(not hasattr(n, 'job'))

#
#
#

def sqr(x, wait=0.0):
    time.sleep(wait)
    return x*x

class _TestPool(BaseTestCase):

    ALLOWED_TYPES = ('processes',)

    def test_apply(self):
        papply = self.pool.apply
        self.assertEqual(papply(sqr, (5,)), apply(sqr, (5,)))
        self.assertEqual(papply(sqr, (), {'x':3}), apply(sqr, (), {'x':3}))

    def test_map(self):
        pmap = self.pool.map
        self.assertEqual(pmap(sqr, range(10)), map(sqr, range(10)))
        self.assertEqual(pmap(sqr, range(100), chunksize=20),
                         map(sqr, range(100)))
        
    def test_async(self):
        res = self.pool.applyAsync(sqr, (7, TIMEOUT1,))
        get = TimingWrapper(res.get)
        self.assertEqual(get(), 49)
        self.assertTimingAlmostEqual(get.elapsed, TIMEOUT1)

    def test_async_timeout(self):
        res = self.pool.applyAsync(sqr, (6, 100))
        get = TimingWrapper(res.get)
        self.assertRaises(processing.TimeoutError, get, timeout=TIMEOUT2)
        self.assertTimingAlmostEqual(get.elapsed, TIMEOUT2)
        # One of the workers will be occupied a long time (probably
        # till the pool gets terminated), but there are other workers
        # so who cares.

    def test_imap(self):
        it = self.pool.imap(sqr, range(10))
        self.assertEqual(list(it), map(sqr, range(10)))

        it = self.pool.imap(sqr, range(10))
        for i in range(10):
            self.assertEqual(it.next(), i*i)
        self.assertRaises(StopIteration, it.next)

        it = self.pool.imap(sqr, range(1000), chunksize=100)
        for i in range(1000):
            self.assertEqual(it.next(), i*i)
        self.assertRaises(StopIteration, it.next)

    def test_imap_unordered(self):
        it = self.pool.imapUnordered(sqr, range(1000))
        self.assertEqual(sorted(it), map(sqr, range(1000)))

        it = self.pool.imapUnordered(sqr, range(1000), chunksize=53)
        self.assertEqual(sorted(it), map(sqr, range(1000)))

    def test_make_pool(self):
        p = processing.Pool(3)
        self.assertEqual(3, len(p._pool))

#
#
#

class _TestZZZDebugInfo(BaseTestCase):

    ALLOWED_TYPES = ('manager',)

    def test_debug_info(self):
        # this gets run after all the other tests for the manager
        # and it tests that there have been no reference leaks for
        # the managers shared objects
        debug = self._debugInfo()
        if debug:
            print debug
        self.assertTrue(not debug)

#
#
#

from processing.managers import (
    BaseManager, BaseProxy, CreatorMethod, RemoteError
    )
    
class FooBar(object):
    def f(self):
        return 'f()'
    def g(self):
        raise ValueError
    def _h(self):
        return '_h()'
    
def baz():
    for i in xrange(10):
        yield i*i

class IteratorProxy(BaseProxy):
    def __iter__(self):
        return self
    def next(self):
        return self._callMethod('next')

class MyManager(BaseManager):
    Foo = CreatorMethod(FooBar)
    Bar = CreatorMethod(FooBar, exposed=('f', '_h'))
    baz = CreatorMethod(baz, proxytype=IteratorProxy)


class _TestMyManager(BaseTestCase):
    
    ALLOWED_TYPES = ('manager',)

    def test_mymanager(self):
        manager = MyManager()
        manager.start()
        
        foo = manager.Foo()
        bar = manager.Bar()
        baz = manager.baz()
        
        foo_methods = [name for name in ('f', 'g', '_h') if hasattr(foo, name)]
        bar_methods = [name for name in ('f', 'g', '_h') if hasattr(bar, name)]
        
        self.assertEqual(foo_methods, ['f', 'g'])
        self.assertEqual(bar_methods, ['f', '_h'])
        
        self.assertEqual(foo.f(), 'f()')
        self.assertRaises(ValueError, foo.g)
        self.assertEqual(foo._callMethod('f'), 'f()')
        self.assertRaises(RemoteError, foo._callMethod, '_h')
        
        self.assertEqual(bar.f(), 'f()')
        self.assertEqual(bar._h(), '_h()')
        self.assertEqual(bar._callMethod('f'), 'f()')
        self.assertEqual(bar._callMethod('_h'), '_h()')
        
        self.assertEqual(list(baz), [i*i for i in range(10)])

        manager.shutdown()
        
#
#
#

_queue = Queue.Queue()

def get_queue():
    return _queue

class QueueManager(BaseManager):
    # manager class used by server process
    get_proxy = CreatorMethod(callable=get_queue, typeid='get_proxy')

class QueueManager2(BaseManager):
    # manager class which specifies the same interface as QueueManager
    get_proxy = CreatorMethod(typeid='get_proxy')


class _TestRemoteManager(BaseTestCase):

    ALLOWED_TYPES = ('manager',)
    
    def _putter(self, address):
        m2 = QueueManager2.fromAddress(address=address, authkey='none')
        queue = m2.get_proxy()
        queue.put('hello world')

    def _test_remote(self):
        m = QueueManager(address=('localhost', 0), authkey='none')
        m.start()
        
        p = self.Process(target=self._putter, args=(m.address,))
        p.start()
        
        m2 = QueueManager2.fromAddress(address=m.address, authkey='none')
        queue = m2.get_proxy()
        self.assertEqual(queue.get(), 'hello world')
        
        # Since queue was not created using m it has no way of knowing
        # when the server process associated with m has been
        # finalized.  Therefore we should take care that queue is
        # finalized before m is finalized.  Otherwise the finalizer
        # for queue can hang for a few seconds (on Windows) while it
        # tries to contact the manager process (which is no longer
        # running) in order to decrement the reference count.
        del queue
        del m

#
#
#

class _TestConnection(BaseTestCase):

    ALLOWED_TYPES = ('processes', 'threads')

    def _echo(self, conn):
        for msg in iter(conn.recvBytes, ''):
            conn.sendBytes(msg)
        conn.close()

    def test_connection(self):
        conn, child_conn = self.Pipe()
        
        p = self.Process(target=self._echo, args=(child_conn,))
        p.setDaemon(True)
        p.start()

        seq = [1, 2.25, None]
        msg = 'hello world'
        longmsg = msg * 10
        arr = array.array('i', range(4))

        if self.TYPE == 'processes':
            self.assertEqual(type(conn.fileno()), int)

        self.assertEqual(conn.send(seq), None)
        self.assertEqual(conn.recv(), seq)

        self.assertEqual(conn.sendBytes(msg), None)
        self.assertEqual(conn.recvBytes(), msg)

        if self.TYPE == 'processes':
            buffer = array.array('i', [0]*10)
            expected = list(arr) + [0] * (10 - len(arr))
            self.assertEqual(conn.sendBytes(arr), None)
            self.assertEqual(conn.recvBytesInto(buffer),
                             len(arr) * buffer.itemsize)
            self.assertEqual(list(buffer), expected)

            buffer = array.array('i', [0]*10)
            expected = [0] * 3 + list(arr) + [0] * (10 - 3 - len(arr))
            self.assertEqual(conn.sendBytes(arr), None)
            self.assertEqual(conn.recvBytesInto(buffer, 3 * buffer.itemsize),
                             len(arr) * buffer.itemsize)
            self.assertEqual(list(buffer), expected)

            buffer = array.array('c', ' ' * 40)
            self.assertEqual(conn.sendBytes(longmsg), None)
            try:
                res = conn.recvBytesInto(buffer)
            except processing.BufferTooShort, e:
                self.assertEqual(e.args, (longmsg,))
            else:
                self.fail('expected BufferTooShort, got %s' % res)

        poll = TimingWrapper(conn.poll)

        self.assertEqual(poll(), False)
        self.assertTimingAlmostEqual(poll.elapsed, 0)

        self.assertEqual(poll(TIMEOUT1), False)
        self.assertTimingAlmostEqual(poll.elapsed, TIMEOUT1)

        conn.send(None)

        self.assertEqual(poll(TIMEOUT1), True)
        self.assertTimingAlmostEqual(poll.elapsed, 0)
        
        self.assertEqual(conn.recv(), None)

        really_big_msg = 'X' * (1024 * 1024 * 16)       # 16 megabytes
        conn.sendBytes(really_big_msg)
        self.assertEqual(conn.recvBytes(), really_big_msg)
        
        conn.sendBytes('')                              # tell child to quit
        child_conn.close()

        if self.TYPE == 'processes':
            self.assertRaises(EOFError, conn.recv)
            self.assertRaises(EOFError, conn.recvBytes)

        p.join()
        
    def test_duplex_false(self):
        reader, writer = self.Pipe(duplex=False)
        self.assertEqual(writer.send(1), None)
        self.assertEqual(reader.recv(), 1)
        if self.TYPE == 'processes':
            self.assertRaises(IOError, writer.recv)
            self.assertRaises(IOError, reader.send, 2)

    def test_spawn_close(self):
        # We test that a pipe connection can be closed by parent
        # process immediately after child is spawned.  On Windows this
        # would have sometimes failed on old versions because
        # child_conn would be closed before the child got a chance to
        # duplicate it.
        conn, child_conn = self.Pipe()
        
        p = self.Process(target=self._echo, args=(child_conn,))
        p.start()
        child_conn.close()    # this might complete before child initializes
        
        conn.sendBytes('hello')
        self.assertEqual(conn.recvBytes(), 'hello')

        conn.sendBytes('')
        conn.close()
        p.join()


class _TestListenerClient(BaseTestCase):

    ALLOWED_TYPES = ('processes', 'threads')

    def _test(self, address):
        conn = self.connection.Client(address)
        conn.send('hello')
        conn.close()

    def _test_listener_client(self):        
        for family in self.connection.families:
            l = self.connection.Listener(family=family)
            p = self.Process(target=self._test, args=(l.address,))
            p.setDaemon(True)
            p.start()
            conn = l.accept()
            self.assertEqual(conn.recv(), 'hello')
            p.join()
            l.close()


class _TestPicklingConnections(BaseTestCase):

    ALLOWED_TYPES = ('processes', 'threads')

    def _listener(self, conn, families):
        for fam in families:
            l = self.connection.Listener(family=fam)
            conn.send(l.address)
            new_conn = l.accept()
            conn.send(new_conn)

        if self.TYPE == 'processes':
            l = socket.socket()
            l.bind(('localhost', 0))
            conn.send(l.getsockname())
            l.listen(1)
            new_conn, addr = l.accept()
            conn.send(new_conn)
        
        conn.recv()

    def _remote(self, conn):
        for (address, msg) in iter(conn.recv, None):
            client = self.connection.Client(address)
            client.send(msg.upper())
            client.close()

        if self.TYPE == 'processes':
            address, msg = conn.recv()
            client = socket.socket()
            client.connect(address)
            client.sendall(msg.upper())
            client.close()

        conn.close()

    def _test_pickling(self):
        if not self.connection.connections_are_picklable:
            return
        
        families = self.connection.families

        lconn, lconn0 = self.Pipe()
        lp = self.Process(target=self._listener, args=(lconn0, families))
        lp.start()
        lconn0.close()

        rconn, rconn0 = self.Pipe()
        rp = self.Process(target=self._remote, args=(rconn0,))
        rp.start()
        rconn0.close()

        for fam in families:
            msg = 'This connection uses family %s' % fam
            address = lconn.recv()
            rconn.send((address, msg))
            new_conn = lconn.recv()
            self.assertEqual(new_conn.recv(), msg.upper())
            
        rconn.send(None)

        if self.TYPE == 'processes':
            msg = 'This connection uses a normal socket'
            address = lconn.recv()
            rconn.send((address, msg))
            new_conn = lconn.recv()
            self.assertEqual(new_conn.recv(100), msg.upper())

        lconn.send(None)

        rconn.close()
        lconn.close()

        lp.join()
        rp.join()

#
#
#

class _TestHeap(BaseTestCase):

    ALLOWED_TYPES = ('processes',)

    def test_heap(self):
        iterations = 5000
        maxblocks = 50
        blocks = []

        # create and destroy lots of blocks of different sizes
        for i in xrange(iterations):
            size = int(random.lognormvariate(0, 1) * 1000)
            b = processing.heap.BufferWrapper(size)
            blocks.append(b)
            if len(blocks) > maxblocks:
                i = random.randrange(maxblocks)
                del blocks[i]

        # get the heap object
        heap = processing.heap.BufferWrapper._heap

        # verify the state of the heap
        all = []
        occupied = 0
        for L in heap._len_to_seq.values():
            for arena, start, stop in L:
                all.append((heap._arenas.index(arena), start, stop,
                            stop-start, 'free'))
        for arena, start, stop in heap._allocated_blocks:
            all.append((heap._arenas.index(arena), start, stop,
                        stop-start, 'occupied'))
            occupied += (stop-start)

        all.sort()

        for i in range(len(all)-1):
            (arena, start, stop) = all[i][:3]
            (narena, nstart, nstop) = all[i+1][:3]
            self.assertTrue((arena != narena and nstart == 0) or
                            (stop == nstart))
            
#
#
#

class _Foo(ctypes.Structure):
    _fields_ = [
        ('x', ctypes.c_int),
        ('y', ctypes.c_double)
        ]

class _TestSharedCTypes(BaseTestCase):

    ALLOWED_TYPES = ('processes',)

    def _double(self, x, y, foo, arr, string):
        x.value *= 2
        y.value *= 2
        foo.x *= 2
        foo.y *= 2
        string.value *= 2
        for i in range(len(arr)):
            arr[i] *= 2

    def test_sharedctypes(self, lock=False):
        from processing.sharedctypes import Value, Array
        
        x = Value('i', 7, lock=lock)
        y = Value(ctypes.c_double, 1.0/3.0, lock=lock)
        foo = Value(_Foo, 3, 2, lock=lock)
        arr = Array('d', range(10), lock=lock)
        string = Array('c', 20, lock=lock)
        string.value = 'hello'

        p = self.Process(target=self._double, args=(x, y, foo, arr, string))
        p.start()
        p.join()

        self.assertEqual(x.value, 14)
        self.assertAlmostEqual(y.value, 2.0/3.0)
        self.assertEqual(foo.x, 6)
        self.assertAlmostEqual(foo.y, 4.0)
        for i in range(10):
            self.assertAlmostEqual(arr[i], i*2)
        self.assertEqual(string.value, 'hellohello')

    def test_synchronize(self):
        self.test_sharedctypes(lock=True)

    def test_copy(self):
        from processing.sharedctypes import Value, copy
        foo = _Foo(2, 5.0)
        bar = copy(foo)
        foo.x = 0
        foo.y = 0
        self.assertEqual(bar.x, 2)
        self.assertAlmostEqual(bar.y, 5.0)

#
#
#

class _TestFinalize(BaseTestCase):

    ALLOWED_TYPES = ('processes',)

    def _test_finalize(self, conn):
        from processing.finalize import Finalize
        from processing.process import _exitFunction

        class Foo(object):
            pass

        a = Foo()
        Finalize(a, conn.send, args=('a',))
        del a           # triggers callback for a

        b = Foo()
        close_b = Finalize(b, conn.send, args=('b',))    
        close_b()       # triggers callback for b
        close_b()       # does nothing because callback has already been called
        del b           # does nothing because callback has already been called

        c = Foo()
        Finalize(c, conn.send, args=('c',))

        d10 = Foo()
        Finalize(d10, conn.send, args=('d10',), exitpriority=1)

        d01 = Foo()
        Finalize(d01, conn.send, args=('d01',), exitpriority=0)
        d02 = Foo()
        Finalize(d02, conn.send, args=('d02',), exitpriority=0)
        d03 = Foo()
        Finalize(d03, conn.send, args=('d03',), exitpriority=0)

        Finalize(None, conn.send, args=('STOP',), exitpriority=-100)

        # call processing's cleanup function then exit process without
        # garbage collecting locals
        _exitFunction()
        conn.close()
        os._exit(0)

    def test_finalize(self):
        conn, child_conn = self.Pipe()
        
        p = self.Process(target=self._test_finalize, args=(child_conn,))
        p.start()
        p.join()

        result = [obj for obj in iter(conn.recv, 'STOP')]
        self.assertEqual(result, ['a', 'b', 'd10', 'd03', 'd02', 'd01'])

#
# Test that from ... import * works for each module
#

class _TestImportStar(BaseTestCase):

    ALLOWED_TYPES = ('processes',)

    def test_import(self):
        modules = (
            'processing', 'processing.connection',
            'processing.finalize', 'processing.forking',
            'processing.heap', 'processing.logger',
            'processing.managers', 'processing.pool',
            'processing.process', 'processing.reduction',
            'processing.sharedctypes', 'processing.synchronize'
            )
        
        for name in modules:
            __import__(name)
            mod = sys.modules[name]
            
            for attr in getattr(mod, '__all__', ()):
                self.assertTrue(
                    hasattr(mod, attr),
                    '%r does not have attribute %r' % (mod, attr)
                    )

#
# Quick test that logging works -- does not test logging output
#

class _TestLogging(BaseTestCase):

    ALLOWED_TYPES = ('processes',)

    def test_enable_logging(self):
        processing.enableLogging(level=processing.SUBWARNING)
        logger = processing.getLogger()
        self.assertTrue(logger is not None)
        logger.debug('this will not be printed')
        logger.info('nor will this')

    def _test_level(self, conn):
        logger = processing.getLogger()
        conn.send(logger.getEffectiveLevel())

    def test_level(self):
        LEVEL1 = 32
        LEVEL2 = 37
        LEVEL3 = 41
        reader, writer = processing.Pipe(duplex=False)
        root_logger = logging.getLogger('')

        processing.enableLogging(level=LEVEL1)
        self.Process(target=self._test_level, args=(writer,)).start()
        self.assertEqual(LEVEL1, reader.recv())

        processing.getLogger().setLevel(LEVEL2)
        self.Process(target=self._test_level, args=(writer,)).start()
        self.assertEqual(LEVEL2, reader.recv())

        self.assertEqual(processing.NOTSET, logging.NOTSET)
        processing.enableLogging(level=logging.NOTSET)
        root_logger.setLevel(LEVEL3)
        self.Process(target=self._test_level, args=(writer,)).start()
        self.assertEqual(LEVEL3, reader.recv())

        processing.enableLogging(level=processing.SUBWARNING)

#
# Functions used to create test cases from the base ones in this module
#

def get_attributes(Source, names):
    d = {}
    for name in names:
        obj = getattr(Source, name)
        if type(obj) == type(get_attributes):
            obj = staticmethod(obj)
        d[name] = obj
    return d

def create_test_cases(Mixin, type):
    result = {}
    glob = globals()
    Type = type[0].upper() + type[1:]

    for name in glob.keys():
        if name.startswith('_Test'):
            base = glob[name]
            if type in base.ALLOWED_TYPES:
                newname = 'With' + Type + name[1:]
                class Temp(base, unittest.TestCase, Mixin):
                    pass
                result[newname] = Temp
                Temp.__name__ = newname
                Temp.__module__ = Mixin.__module__
    return result

#
# Create test cases
#

class ProcessesMixin(object):
    TYPE = 'processes'
    Process = processing.Process
    locals().update(get_attributes(processing, (
        'Queue', 'Lock', 'RLock', 'Semaphore', 'BoundedSemaphore',
        'Condition', 'Event', 'Value', 'Array',
        'currentProcess', 'activeChildren', 'Pipe', 'connection'
        )))

testcases_processes = create_test_cases(ProcessesMixin, type='processes')
globals().update(testcases_processes)


class ManagerMixin(object):
    TYPE = 'manager'
    Process = processing.Process
    manager = object.__new__(processing.managers.SyncManager)
    locals().update(get_attributes(manager, (
        'Queue', 'Lock', 'RLock', 'Semaphore', 'BoundedSemaphore',
        'Condition', 'Event', 'Value', 'Array', 'list', 'dict',
        'Namespace', '_debugInfo'
        )))

testcases_manager = create_test_cases(ManagerMixin, type='manager')
globals().update(testcases_manager)


class ThreadsMixin(object):
    TYPE = 'threads'
    Process = processing.dummy.Process
    locals().update(get_attributes(processing.dummy, (
        'Queue', 'Lock', 'RLock', 'Semaphore', 'BoundedSemaphore',
        'Condition', 'Event', 'Value', 'Array', 'currentProcess',
        'activeChildren', 'Pipe', 'connection', 'dict', 'list',
        'Namespace'
        )))

testcases_threads = create_test_cases(ThreadsMixin, type='threads')
globals().update(testcases_threads)

#
#
#

def test_main(run=None):
    if run is None:
        from test.test_support import run_suite as run

    ProcessesMixin.pool = processing.Pool(4)
    ManagerMixin.manager.__init__()
    ManagerMixin.manager.start()

    testcases = (
        sorted(testcases_processes.values(), key=lambda tc:tc.__name__) +
        sorted(testcases_threads.values(), key=lambda tc:tc.__name__) +
        sorted(testcases_manager.values(), key=lambda tc:tc.__name__)
        )

    loadTestsFromTestCase = unittest.defaultTestLoader.loadTestsFromTestCase
    suite = unittest.TestSuite(loadTestsFromTestCase(tc) for tc in testcases)
    run(suite)
    
    ManagerMixin.manager.shutdown()
    ProcessesMixin.pool.terminate()
    del ProcessesMixin.pool

def main():
    test_main(unittest.TextTestRunner(verbosity=2).run)

if __name__ == '__main__':
    main()



#
# Module implementing queues
#
# processing/queue.py
#
# Copyright (c) 2006-2008, R Oudkerk --- see COPYING.txt
#

__all__ = ['Queue', 'SimpleQueue']

import sys
import os
import threading
import collections
import time
import atexit
import weakref

from Queue import Empty, Full

from processing import _processing, Pipe, currentProcess
from processing.synchronize import Lock, BoundedSemaphore
from processing.logger import debug, subWarning
from processing.finalize import Finalize
from processing.process import _exitFunction, _registerAfterFork
from processing.forking import assertSpawning

#
# Cleanup function of `processing` should run before that of `threading`
#

atexit._exithandlers.remove((_exitFunction, (), {}))
atexit._exithandlers.append((_exitFunction, (), {}))

#
# Queue type using a pipe, buffer and thread
#

class Queue(object):

    def __init__(self, maxsize=0):
        reader, writer = Pipe(duplex=False)
        rlock = Lock()
        if sys.platform == 'win32':
            wlock = None
        else:
            wlock = Lock()
        if maxsize < 0:
            maxsize = 0
        if maxsize == 0:
            sem = None
        else:
            sem = BoundedSemaphore(maxsize)

        state = maxsize, reader, writer, rlock, wlock, sem, os.getpid()
        self.__setstate__(state)
        
        if sys.platform != 'win32':
            _registerAfterFork(self, Queue._afterFork)

    def __getstate__(self):
        assertSpawning(self)
        return self._state

    def __setstate__(self, state):
        (self._maxsize, self._reader, self._writer,
         self._rlock, self._wlock, self._sem, self._opid) = self._state = state
        self._send = self._writer.send
        self._recv = self._reader.recv
        self._poll = self._reader.poll
        self._afterFork()

    def _afterFork(self):
        debug('Queue._afterFork()')
        self._notempty = threading.Condition(threading.Lock())
        self._buffer = collections.deque()
        self._thread = None
        self._jointhread = None
        self._joincancelled = False
        self._closed = False
        self._close = None
        
    def put(self, obj, block=True, timeout=None):
        assert not self._closed
        if self._sem is not None:
            if not self._sem.acquire(block, timeout):
                raise Full

        self._notempty.acquire()
        try:
            if self._thread is None:
                self._startThread()
            self._buffer.append(obj)
            self._notempty.notify()
        finally:
            self._notempty.release()

    def putMany(self, iterable):
        assert not self._closed
        assert self._maxsize == 0

        self._notempty.acquire()
        try:
            if self._thread is None:
                self._startThread()
            self._buffer.extend(iterable)
            self._notempty.notify()
        finally:
            self._notempty.release()

    def get(self, block=True, timeout=None):
        if block and timeout is None:
            self._rlock.acquire()
            try:
                res = self._recv()
                if self._sem:
                    self._sem.release()
                return res
            finally:
                self._rlock.release()
                
        else:
            if block:
                deadline = time.time() + timeout
            if not self._rlock.acquire(block, timeout):
                raise Empty
            try:
                if not self._poll(block and (deadline-time.time()) or 0.0):
                    raise Empty
                res = self._recv()
                if self._sem:
                    self._sem.release()
                return res
            finally:
                self._rlock.release()

    def empty(self):
        # Even more unreliable than Queue.Queue.empty(): True can be
        # returned when enqueued items are buffered but none are
        # yet in the pipe
        return not self._poll()

    def full(self):
        if self._sem:
            if self._sem.acquire(False):
                self._sem.release()
                return False
            return True
        else:
            return False

    def getNoWait(self):
        return self.get(False)

    def putNoWait(self, obj):
        return self.put(obj, False)

    def close(self):
        self._closed = True
        self._reader.close()
        if self._close:
            self._close()

    def joinThread(self):
        debug('Queue.joinThread()')
        assert self._closed
        if self._jointhread:
            self._jointhread()
    
    def cancelJoin(self):
        debug('Queue.cancelJoin()')
        self._joincancelled = True
        try:
            self._jointhread.cancel()
        except AttributeError:
            pass

    def _startThread(self):
        debug('Queue._startThread()')
        
        # Start thread which transfers data from buffer to pipe
        self._buffer.clear()
        self._thread = threading.Thread(
            target=Queue._feed,
            args=(self._buffer, self._notempty, self._send,
                  self._wlock, self._writer.close),
            name='QueueFeederThread'
            )
        self._thread.setDaemon(True)

        debug('doing self._thread.start()')
        self._thread.start()
        debug('... done self._thread.start()')

        # On process exit we will wait for data to be flushed to pipe.
        #
        # However, if this process created the queue then all
        # processes which use the queue will be descendants of this
        # process.  Therefore waiting for the queue to be flushed
        # is pointless once all the child processes have been joined.
        created_by_this_process = (self._opid == os.getpid())
        if not self._joincancelled and not created_by_this_process:
            self._jointhread = Finalize(
                self._thread, Queue._finalizeJoin,
                [weakref.ref(self._thread)],
                exitpriority=-5
                )
            
        # Send sentinel to the thread queue object when garbage collected
        self._close = Finalize(
            self, Queue._finalizeClose,
            [self._buffer, self._notempty],
            exitpriority=10
            )
        
    @staticmethod
    def _finalizeJoin(twr):
        debug('joining queue thread')
        thread = twr()
        if thread is not None:
            thread.join()
            debug('... queue thread joined')
        else:
            debug('... queue thread already dead')
            
    @staticmethod 
    def _finalizeClose(buffer, notempty):
        debug('telling queue thread to quit')
        notempty.acquire()
        try:
            buffer.append(_sentinel)
            notempty.notify()
        finally:
            notempty.release()              

    @staticmethod
    def _feed(buffer, notempty, send, writelock, close):
        debug('starting thread to feed data to pipe')

        nacquire = notempty.acquire
        nrelease = notempty.release
        nwait = notempty.wait
        bpopleft = buffer.popleft
        sentinel = _sentinel
        if sys.platform != 'win32':
            wacquire = writelock.acquire
            wrelease = writelock.release
        else:
            wacquire = None
        
        try:
            while 1:
                nacquire()
                try:
                    if not buffer:
                        nwait()
                finally:
                    nrelease()
                try:
                    while 1:
                        obj = bpopleft()
                        if obj is sentinel:
                            debug('feeder thread got sentinel -- exiting')
                            close()
                            return

                        if wacquire is None:
                            send(obj)
                        else:
                            wacquire()
                            try:
                                send(obj)
                            finally:
                                wrelease()                        
                except IndexError:
                    pass
        except Exception, e:
            # Since this runs in a daemon thread the objects it uses
            # may be become unusable while the process is cleaning up.
            # We ignore errors which happen after the process has
            # started to cleanup.
            if currentProcess()._exiting:
                subWarning('error in queue thread: %s', e)
            else:
                raise
            
    get_nowait = getNoWait
    put_nowait = putNoWait

    # deprecated
    putmany = putMany        
    jointhread = joinThread
    canceljoin = cancelJoin

_sentinel = object()

#
# Simplified Queue type -- really just a locked pipe
#

class SimpleQueue(object):

    def __init__(self):
        reader, writer = Pipe(duplex=False)
        if sys.platform == 'win32':
            state = reader, writer, Lock(), None
        else:
            state = reader, writer, Lock(), Lock()
        self.__setstate__(state)

    def empty(self):
        return not self._reader.poll()

    def __getstate__(self):
        assertSpawning(self)
        return self._state

    def __setstate__(self, state):
        (self._reader, self._writer, self._rlock, self._wlock) \
                       = self._state = state

        recv = self._reader.recv
        racquire, rrelease = self._rlock.acquire, self._rlock.release
        def get():
            racquire()
            try:
                return recv()
            finally:
                rrelease()
        self.get = get

        if self._wlock is None:
            # writes to a message oriented win32 pipe are atomic
            self.put = self._writer.send
        else:
            send = self._writer.send
            wacquire, wrelease = self._wlock.acquire, self._wlock.release
            def put(obj):
                wacquire()
                try:
                    return send(obj)
                finally:
                    wrelease()
            self.put = put



#
# Module providing the `Pool` class for managing a process pool
#
# processing/pool.py
#
# Copyright (c) 2007-2008, R Oudkerk --- see COPYING.txt
#

__all__ = ['Pool']

#
# Imports
#

import processing
import threading
import Queue
import itertools
import collections
import time

from processing import Process
from processing.logger import debug
from processing.finalize import Finalize
from processing.queue import SimpleQueue

#
# Constants representing the state of a pool
#

RUN = 0
CLOSE = 1
TERMINATE = 2

#
# Miscellaneous
#

newJobId = itertools.count().next

def mapstar(args):
    return map(*args)

#
# Code run by worker processes
#

def worker(inqueue, outqueue, initializer=None, initargs=()):
    put = outqueue.put

    if initializer is not None:
        initializer(*initargs)

    for job, i, func, args, kwds in iter(inqueue.get, None):
        try:
            result = (True, func(*args, **kwds))
        except Exception, e:
            result = (False, e)
        put((job, i, result))

    debug('worker got sentinel -- exiting')
    
#
# Class representing a process pool
#

class Pool(object):
    '''
    Class which supports an async version of the `apply()` builtin
    '''
    def __init__(self, processes=None, initializer=None, initargs=()):
        self._inqueue = SimpleQueue()
        self._outqueue = SimpleQueue()
        self._taskqueue = Queue.Queue()
        self._cache = {}
        self._state = RUN

        if processes is None:
            try:
                processes = processing.cpuCount()
            except NotImplementedError:
                processes = 1
            
        self._pool = [
            Process(target=worker, args=(self._inqueue, self._outqueue,
                                         initializer, initargs))
            for i in range(processes)
            ]
        
        for i, w in enumerate(self._pool):
            w.setName('PoolWorker-' + ':'.join(map(str, w._identity)))
            w.start()
                    
        self._task_handler = threading.Thread(
            target=Pool._handleTasks,
            args=(self._taskqueue, self._inqueue, self._outqueue, self._pool)
            )
        self._task_handler.setDaemon(True)
        self._task_handler._state = RUN
        self._task_handler.start()

        self._result_handler = threading.Thread(
            target=Pool._handleResults,
            args=(self._outqueue, self._cache)
            )
        self._result_handler.setDaemon(True)
        self._result_handler._state = RUN
        self._result_handler.start()

        self._terminate = Finalize(
            self, Pool._terminatePool,
            args=(self._taskqueue, self._inqueue, self._outqueue,
                  self._cache, self._pool, self._task_handler,
                  self._result_handler),
            exitpriority=5
            )

    def apply(self, func, args=(), kwds={}):
        '''
        Equivalent of `apply()` builtin
        '''
        assert self._state == RUN
        return self.applyAsync(func, args, kwds).get()

    def map(self, func, iterable, chunksize=None):
        '''
        Equivalent of `map()` builtin
        '''
        assert self._state == RUN
        return self.mapAsync(func, iterable, chunksize).get()

    def imap(self, func, iterable, chunksize=1):
        '''
        Equivalent of `itertool.imap()` -- can be MUCH slower than `Pool.map()`
        '''
        assert self._state == RUN
        if chunksize == 1:
            result = IMapIterator(self._cache)
            self._taskqueue.put((((result._job, i, func, (x,), {})
                          for i, x in enumerate(iterable)), result._setLength))
            return result
        else:
            assert chunksize > 1
            task_batches = Pool._getTasks(func, iterable, chunksize)
            result = IMapIterator(self._cache)
            self._taskqueue.put((((result._job, i, mapstar, (x,), {})
                      for i, x in enumerate(task_batches)), result._setLength))
            return (item for chunk in result for item in chunk)

    def imapUnordered(self, func, iterable, chunksize=1):
        '''
        Like `imap()` method but ordering of results is arbitrary
        '''
        assert self._state == RUN
        if chunksize == 1:
            result = IMapUnorderedIterator(self._cache)
            self._taskqueue.put((((result._job, i, func, (x,), {})
                          for i, x in enumerate(iterable)), result._setLength))
            return result
        else:
            assert chunksize > 1
            task_batches = Pool._getTasks(func, iterable, chunksize)
            result = IMapUnorderedIterator(self._cache)
            self._taskqueue.put((((result._job, i, mapstar, (x,), {})
                      for i, x in enumerate(task_batches)), result._setLength))
            return (item for chunk in result for item in chunk)
            
    def applyAsync(self, func, args=(), kwds={}, callback=None):
        '''
        Asynchronous equivalent of `apply()` builtin
        '''
        assert self._state == RUN
        result = ApplyResult(self._cache, callback)
        self._taskqueue.put(([(result._job, None, func, args, kwds)], None))
        return result

    def mapAsync(self, func, iterable, chunksize=None, callback=None):
        '''
        Asynchronous equivalent of `map()` builtin
        '''
        assert self._state == RUN
        if not hasattr(iterable, '__len__'):
            iterable = list(iterable)
        
        if chunksize is None:
            chunksize, extra = divmod(len(iterable), len(self._pool) * 4)
            if extra:
                chunksize += 1
                
        task_batches = Pool._getTasks(func, iterable, chunksize)
        result = MapResult(self._cache, chunksize, len(iterable), callback)
        self._taskqueue.put((((result._job, i, mapstar, (x,), {})
                              for i, x in enumerate(task_batches)), None))
        return result

    @staticmethod
    def _handleTasks(taskqueue, inqueue, outqueue, pool):
        thread = threading.currentThread()
        put = inqueue._writer.send

        for taskseq, setLength in iter(taskqueue.get, None):
            i = -1
            for i, task in enumerate(taskseq):
                if thread._state:
                    debug('task handler found thread._state != RUN')
                    break
                put(task)
            else:
                if setLength:
                    debug('doing setLength()')
                    setLength(i+1)
                continue
            break
        else:
            debug('task handler got sentinel')
            
        # tell result handler to finish when cache is empty
        outqueue.put(None)

        # tell workers there is no more work
        debug('task handler sending sentinel to workers')
        for p in pool:
            put(None)

        debug('task handler exiting')

    @staticmethod
    def _handleResults(outqueue, cache):
        thread = threading.currentThread()
        get = outqueue._reader.recv

        for job, i, obj in iter(get, None):
            if thread._state:
                assert thread._state == TERMINATE
                debug('result handler found thread._state=TERMINATE')
                return
            try:
                cache[job]._set(i, obj)
            except KeyError:
                pass
        else:
            debug('result handler got sentinel')

        while cache and thread._state != TERMINATE:
            item = get()
            if item is None:
                debug('result handler ignoring extra sentinel')
                continue
            job, i, obj = item
            try:
                cache[job]._set(i, obj)
            except KeyError:
                pass

        debug('result handler exiting: len(cache)=%s, thread._state=%s',
              len(cache), thread._state)

    @staticmethod
    def _getTasks(func, it, size):
        it = iter(it)
        while 1:
            x = tuple(itertools.islice(it, size))
            if not x:
                return
            yield (func, x)

    def __reduce__(self):
        raise NotImplementedError, \
              'pool objects cannot be passed between processes or pickled'
    
    def close(self):
        debug('closing pool')
        self._state = CLOSE
        self._taskqueue.put(None)

    def terminate(self):
        debug('terminating pool')
        self._state = TERMINATE
        self._terminate()

    def join(self):
        debug('joining pool')
        assert self._state in (CLOSE, TERMINATE)
        self._task_handler.join()
        self._result_handler.join()
        for p in self._pool:
            p.join()

    @staticmethod
    def _terminatePool(taskqueue, inqueue, outqueue, cache, pool,
                        task_handler, result_handler):
        debug('finalizing pool')
        
        if not result_handler.isAlive():
            debug('result handler already finished -- no need to terminate')
            return
        
        cache = {}
        task_handler._state = TERMINATE
        result_handler._state = TERMINATE

        debug('sending sentinels')
        taskqueue.put(None)
        outqueue.put(None)
        
        debug('getting read lock on inqueue')
        inqueue._rlock.acquire()

        debug('terminating workers')
        for p in pool:
            p.terminate()

        if task_handler.isAlive():
            debug('removing tasks from inqueue until task handler finished')
            while task_handler.isAlive() and inqueue._reader.poll():
                inqueue._reader.recv()
                time.sleep(0)

        debug('joining result handler')
        result_handler.join()
        debug('joining task handler')
        task_handler.join()
        debug('joining pool workers')
        for p in pool:
            p.join()
        debug('closing connections')
        inqueue._reader.close()
        outqueue._reader.close()
        inqueue._writer.close()
        outqueue._writer.close()

    # deprecated
    apply_async = applyAsync
    map_async = mapAsync
    imap_unordered = imapUnordered

#
# Class whose instances are returned by `Pool.applyAsync()`
#

class ApplyResult(object):

    def __init__(self, cache, callback):
        self._cond = threading.Condition(threading.Lock())
        self._job = newJobId()
        self._cache = cache
        self._ready = False
        self._callback = callback
        cache[self._job] = self
        
    def ready(self):
        return self._ready
    
    def successful(self):
        assert self._ready
        return self._success
    
    def wait(self, timeout=None):
        self._cond.acquire()
        try:
            if not self._ready:
                self._cond.wait(timeout)
        finally:
            self._cond.release()

    def get(self, timeout=None):
        self.wait(timeout)
        if not self._ready:
            raise processing.TimeoutError
        if self._success:
            return self._value
        else:
            raise self._value

    def _set(self, i, obj):
        self._success, self._value = obj
        if self._callback and self._success:
            self._callback(self._value)
        self._cond.acquire()
        try:
            self._ready = True
            self._cond.notify()
        finally:
            self._cond.release()
        del self._cache[self._job]

#
# Class whose instances are returned by `Pool.mapAsync()`
#

class MapResult(ApplyResult):
    
    def __init__(self, cache, chunksize, length, callback):
        ApplyResult.__init__(self, cache, callback)
        self._success = True
        self._value = [None] * length
        self._chunksize = chunksize
        if chunksize <= 0:
            self._number_left = 0
            self._ready = True
        else:
            self._number_left = length//chunksize + bool(length % chunksize)
        
    def _set(self, i, (success, result)):
        if success:
            self._value[i*self._chunksize:(i+1)*self._chunksize] = result
            self._number_left -= 1
            if self._number_left == 0:
                if self._callback:
                    self._callback(self._value)
                del self._cache[self._job]
                self._cond.acquire()
                try:
                    self._ready = True
                    self._cond.notify()
                finally:
                    self._cond.release()

        else:
            self._success = False
            self._value = result
            del self._cache[self._job]
            self._cond.acquire()
            try:
                self._ready = True
                self._cond.notify()
            finally:
                self._cond.release()

#
# Class whose instances are returned by `Pool.imap()`
#

class IMapIterator(object):

    def __init__(self, cache):
        self._cond = threading.Condition(threading.Lock())
        self._job = newJobId()
        self._cache = cache
        self._items = collections.deque()
        self._index = 0
        self._length = None
        self._unsorted = {}
        cache[self._job] = self
        
    def __iter__(self):
        return self
    
    def next(self, timeout=None):
        self._cond.acquire()
        try:
            try:
                item = self._items.popleft()
            except IndexError:
                if self._index == self._length:
                    raise StopIteration
                self._cond.wait(timeout)
                try:
                    item = self._items.popleft()
                except IndexError:
                    if self._index == self._length:
                        raise StopIteration
                    raise processing.TimeoutError
        finally:
            self._cond.release()

        success, value = item
        if success:
            return value
        raise value
    
    def _set(self, i, obj):
        self._cond.acquire()
        try:
            if self._index == i:
                self._items.append(obj)
                self._index += 1
                while self._index in self._unsorted:
                    obj = self._unsorted.pop(self._index)
                    self._items.append(obj)
                    self._index += 1
                self._cond.notify()
            else:
                self._unsorted[i] = obj
                
            if self._index == self._length:
                del self._cache[self._job]
        finally:
            self._cond.release()
            
    def _setLength(self, length):
        self._cond.acquire()
        try:
            self._length = length
            if self._index == self._length:
                self._cond.notify()
                del self._cache[self._job]
        finally:
            self._cond.release()

#
# Class whose instances are returned by `Pool.imapUnordered()`
#

class IMapUnorderedIterator(IMapIterator):

    def _set(self, i, obj):
        self._cond.acquire()
        try:
            self._items.append(obj)
            self._index += 1
            self._cond.notify()
            if self._index == self._length:
                del self._cache[self._job]
        finally:
            self._cond.release()



#!/usr/bin/env python

import re
import os
import sys
import glob
# drop support for older python
unsupported = None
if sys.version_info < (2, 7):
    unsupported = 'Versions of Python before 2.7 are not supported'
elif (3, 0) <= sys.version_info < (3, 5):
    unsupported = 'Versions of Python before 3.5 are not supported'
if unsupported:
    raise ValueError(unsupported)
stable_version = '0.70.11.1'
pymajor,pyminor = sys.version_info[:2]
pkgdir = 'py%s.%s' % (pymajor,pyminor)
if sys.version_info >= (2, 6):
    pkgname = 'multiprocess'
else: # (2, 5)
    pkgname = 'processing'  #XXX: oddity, due to lazyness at the moment
# if sys.version is higher than explicitly supported, try the latest version
ver = float(pkgdir[2:])
HERE = os.path.dirname(os.path.abspath(__file__))
while not os.path.exists(os.path.join(HERE,'py%s' % ver)):
    ver -= 0.1
if 'py%s' % ver != pkgdir:
    msg = 'Warning: Python %s is not currently supported, reverting to %s'
    print(msg % (pkgdir[2:],ver))
    pkgdir = 'py%s' % ver
srcdir = '%s/Modules/_%s' % (pkgdir, pkgname)
libdir = '%s/%s' % (pkgdir, pkgname)

try:
    from setuptools import setup, Extension, find_packages
    has_setuptools = True
except ImportError:
    from distutils.core import setup, Extension  # noqa
    find_packages = lambda **kwds: [pkgname, pkgname+'.dummy', pkgname+'.tests']
    has_setuptools = False
from distutils import sysconfig
from distutils.errors import CCompilerError, DistutilsExecError, \
                                             DistutilsPlatformError

ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)
if sys.platform == 'win32' and sys.version_info >= (2, 6):
    # distutils.msvc9compiler can raise IOError if the compiler is missing
    ext_errors += (IOError, )

is_jython = sys.platform.startswith('java')
is_pypy = hasattr(sys, 'pypy_version_info')
is_py3k = sys.version_info[0] == 3
lt_py33 = sys.version_info < (3, 3)

BUILD_WARNING = """

-----------------------------------------------------------------------
WARNING: The C extensions could not be compiled
-----------------------------------------------------------------------

Maybe you do not have a C compiler installed on this system?
The reason was:
%s

This is just a warning as most of the functionality will work even
without the updated C extension.  It will simply fallback to the
built-in _multiprocessing module.  Most notably you will not be able to use
FORCE_EXECV on POSIX systems.  If this is a problem for you then please
install a C compiler or fix the error(s) above.
-----------------------------------------------------------------------

"""

# -*- extra config (setuptools) -*-
if has_setuptools:
    extras = dict(install_requires=['dill>=0.3.3'])
else:
    extras = dict()

# -*- Distribution Meta -*-
here = os.path.abspath(os.path.dirname(__file__))
meta_fh = open(os.path.join(here, '%s/__init__.py' % libdir))
try:
    meta = {}
    for line in meta_fh:
        if line.startswith('__version__'):
            version = line.split()[-1].strip("'").strip('"')
            break
    meta['version'] = version
finally:
    meta_fh.close()

#
# Macros and libraries
#
#   The `macros` dict determines the macros that will be defined when
#   the C extension is compiled.  Each value should be either 0 or 1.
#   (An undefined macro is assumed to have value 0.)  `macros` is only
#   used on Unix platforms.
#
#   The `libraries` dict determines the libraries to which the C
#   extension will be linked.  This should probably be either `['rt']`
#   if you need `librt` or else `[]`.
#
# Meaning of macros
#
#   HAVE_SEM_OPEN
#     Set this to 1 if you have `sem_open()`.  This enables the use of
#     posix named semaphores which are necessary for the
#     implementation of the synchronization primitives on Unix.  If
#     set to 0 then the only way to create synchronization primitives
#     will be via a manager (e.g. "m = Manager(); lock = m.Lock()").
#     
#   HAVE_SEM_TIMEDWAIT
#     Set this to 1 if you have `sem_timedwait()`.  Otherwise polling
#     will be necessary when waiting on a semaphore using a timeout.
#     
#   HAVE_FD_TRANSFER
#     Set this to 1 to compile functions for transferring file
#     descriptors between processes over an AF_UNIX socket using a
#     control message with type SCM_RIGHTS.  On Unix the pickling of 
#     of socket and connection objects depends on this feature.
#
#     If you get errors about missing CMSG_* macros then you should
#     set this to 0.
# 
#   HAVE_BROKEN_SEM_GETVALUE
#     Set to 1 if `sem_getvalue()` does not work or is unavailable.
#     On Mac OSX it seems to return -1 with message "[Errno 78]
#     Function not implemented".
#
#   HAVE_BROKEN_SEM_UNLINK
#     Set to 1 if `sem_unlink()` is unnecessary.  For some reason this
#     seems to be the case on Cygwin where `sem_unlink()` is missing
#     from semaphore.h.
#

if sys.platform == 'win32':  # Windows
    macros = dict()
    libraries = ['ws2_32']
elif sys.platform.startswith('darwin'):  # Mac OSX
    macros = dict(
        HAVE_SEM_OPEN=1,
        HAVE_SEM_TIMEDWAIT=0,
        HAVE_FD_TRANSFER=1,
        HAVE_BROKEN_SEM_GETVALUE=1
        )
    libraries = []
elif sys.platform.startswith('cygwin'):  # Cygwin
    macros = dict(
        HAVE_SEM_OPEN=1,
        HAVE_SEM_TIMEDWAIT=1,
        HAVE_FD_TRANSFER=0,
        HAVE_BROKEN_SEM_UNLINK=1
        )
    libraries = []
elif sys.platform in ('freebsd4', 'freebsd5', 'freebsd6'):
    # FreeBSD's P1003.1b semaphore support is very experimental
    # and has many known problems. (as of June 2008)
    macros = dict(                  # FreeBSD 4-6
        HAVE_SEM_OPEN=0,
        HAVE_SEM_TIMEDWAIT=0,
        HAVE_FD_TRANSFER=1,
        )
    libraries = []
elif re.match('^(gnukfreebsd(8|9|10|11)|freebsd(7|8|9|10))', sys.platform):
    macros = dict(                  # FreeBSD 7+ and GNU/kFreeBSD 8+
        HAVE_SEM_OPEN=bool(
            sysconfig.get_config_var('HAVE_SEM_OPEN') and not
            bool(sysconfig.get_config_var('POSIX_SEMAPHORES_NOT_ENABLED'))
        ),
        HAVE_SEM_TIMEDWAIT=1,
        HAVE_FD_TRANSFER=1,
    )
    libraries = []
elif sys.platform.startswith('openbsd'):
    macros = dict(                  # OpenBSD
        HAVE_SEM_OPEN=0,            # Not implemented
        HAVE_SEM_TIMEDWAIT=0,
        HAVE_FD_TRANSFER=1,
    )
    libraries = []
else:                                   # Linux and other unices
    macros = dict(
        HAVE_SEM_OPEN=1,
        HAVE_SEM_TIMEDWAIT=1,
        HAVE_FD_TRANSFER=1,
    )
    libraries = ['rt']

if sys.platform == 'win32':
    multiprocessing_srcs = [
        '%s/%s.c' % (srcdir, pkgname),
        '%s/semaphore.c' % srcdir,
    ]
    if lt_py33:
        multiprocessing_srcs += [
            '%s/pipe_connection.c' % srcdir,
            '%s/socket_connection.c' % srcdir,
            '%s/win32_functions.c' % srcdir,
        ]
else:
    multiprocessing_srcs = [ '%s/%s.c' % (srcdir, pkgname) ]
    if lt_py33:
        multiprocessing_srcs.append('%s/socket_connection.c' % srcdir)

    if macros.get('HAVE_SEM_OPEN', False):
        multiprocessing_srcs.append('%s/semaphore.c' % srcdir)

long_description = \
'''-----------------------------------------------------------------
multiprocess: better multiprocessing and multithreading in python
-----------------------------------------------------------------

About Multiprocess
====================

``multiprocess`` is a fork of ``multiprocessing``, and is developed as part of ``pathos``: 
https://github.com/uqfoundation/pathos

``multiprocessing`` is a package for the Python language which supports the
spawning of processes using the API of the standard library's
``threading`` module. ``multiprocessing`` has been distributed in the standard
library since python 2.6.

Features:

    - Objects can be transferred between processes using pipes or multi-producer/multi-consumer queues.
    - Objects can be shared between processes using a server process or (for simple data) shared memory.
    - Equivalents of all the synchronization primitives in ``threading`` are available.
    - A ``Pool`` class makes it easy to submit tasks to a pool of worker processes.


``multiprocess`` is part of ``pathos``,  a python framework for heterogeneous computing.
``multiprocess`` is in active development, so any user feedback, bug reports, comments,
or suggestions are highly appreciated.  A list of issues is located at https://github.com/uqfoundation/multiprocess/issues, with a legacy list maintained at https://uqfoundation.github.io/pathos-issues.html.

NOTE: A C compiler is required to build the included extension module. For python 3.3 and above, a C compiler is suggested, but not required.


Major Changes
==============

    - enhanced serialization, using ``dill``


Current Release
===============

This documentation is for version ``multiprocess-%(thisver)s`` (a fork of ``multiprocessing-0.70a1``).

The latest released version of ``multiprocess`` is available from::

    https://pypi.org/project/multiprocess

``Multiprocessing`` is distributed under a BSD license.


Development Version
===================

You can get the latest development version with all the shiny new features at::

    https://github.com/uqfoundation

If you have a new contribution, please submit a pull request.


Installation
============

``multiprocess`` is packaged to install from source, so you must
download the tarball, unzip, and run the installer::

    [download]
    $ tar -xvzf multiprocess-%(relver)s.tgz
    $ cd multiprocess-%(relver)s
    $ python setup.py build
    $ python setup.py install

You will be warned of any missing dependencies and/or settings
after you run the "build" step above.

Alternately, ``multiprocess`` can be installed with ``pip`` or ``easy_install``::

    $ pip install multiprocess

NOTE: A C compiler is required to build the included extension module from source. For python 3.3 and above, a C compiler is suggested, but not required. Binary installs do not require a C compiler.


Requirements
============

``multiprocess`` requires::

    - ``python``, **version == 2.7** or **version >= 3.5**
    - ``dill``, **version >= 0.3.3**

Optional requirements::

    - ``setuptools``, **version >= 0.6**


More Information
================

Probably the best way to get started is to look at the documentation at
http://multiprocess.rtfd.io. See ``multiprocess.examples`` for a set of example
scripts. You can also run the test suite with ``python -m multiprocess.tests``.
Please feel free to submit a ticket on github, or ask a question on
stackoverflow (**@Mike McKerns**).  If you would like to share how you use
``multiprocess`` in your work, please post send an email
(to **mmckerns at uqfoundation dot org**).


Citation
========

If you use ``multiprocess`` to do research that leads to publication, we ask that you
acknowledge use of ``multiprocess`` by citing the following in your publication::

    M.M. McKerns, L. Strand, T. Sullivan, A. Fang, M.A.G. Aivazis,
    "Building a framework for predictive science", Proceedings of
    the 10th Python in Science Conference, 2011;
    http://arxiv.org/pdf/1202.1056

    Michael McKerns and Michael Aivazis,
    "pathos: a framework for heterogeneous computing", 2010- ;
    https://uqfoundation.github.io/pathos.html

Please see https://uqfoundation.github.io/pathos.html or
http://arxiv.org/pdf/1202.1056 for further information.
''' % {'relver': stable_version, 'thisver': stable_version}
#long_description = open(os.path.join(HERE, 'README.md')).read()
#long_description += """
#
#===========
#Changes
#===========
#
#"""
#long_description += open(os.path.join(HERE, 'CHANGES.txt')).read()
#if not is_py3k:
#    long_description = long_description.encode('ascii', 'replace')

# -*- Installation Requires -*-
#py_version = sys.version_info
#is_jython = sys.platform.startswith('java')
#is_pypy = hasattr(sys, 'pypy_version_info')

#def strip_comments(l):
#    return l.split('#', 1)[0].strip()
#
#def reqs(f):
#    return list(filter(None, [strip_comments(l) for l in open(
#        os.path.join(os.getcwd(), 'requirements', f)).readlines()]))
#
#if py_version[0] == 3:
#    tests_require = reqs('test3.txt')
#else:
#    tests_require = reqs('test.txt')


def _is_build_command(argv=sys.argv, cmds=('install', 'build', 'bdist')):
    for arg in argv:
        if arg.startswith(cmds):
            return arg


def run_setup(with_extensions=True):
    extensions = []
    if with_extensions:
        extensions = [
            Extension(
                '_%s' % pkgname,
                sources=multiprocessing_srcs,
                define_macros=list(macros.items()),
                libraries=libraries,
                include_dirs=[srcdir],
                depends=glob.glob('%s/*.h' % srcdir) + ['setup.py'],
            ),
        ]
    packages = find_packages(
        where=pkgdir,
        exclude=['ez_setup', 'examples', 'doc',],
        )
    config = dict(
        name='multiprocess',
        version=meta['version'],
        description=('better multiprocessing and multithreading in python'),
        long_description=long_description,
        packages=packages,
        ext_modules=extensions,
        author='Mike McKerns',
        maintainer='Mike McKerns',
        url='https://pypi.org/project/multiprocess',
        download_url='https://github.com/uqfoundation/multiprocess/releases/download/multiprocess-%s/multiprocess-%s.tar.gz' % (stable_version, stable_version),
        zip_safe=False,
        license='BSD',
        package_dir={'': pkgdir},
#       tests_require=tests_require,
#       test_suite='nose.collector',
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: BSD License',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 3',
            'Topic :: Scientific/Engineering',
            'Topic :: Software Development',
        ],
        **extras
    )
    setup(**config)

try:
    run_setup(not (is_jython or is_pypy) and lt_py33)
except BaseException:
    if _is_build_command(sys.argv):
        import traceback
        msg = BUILD_WARNING % '\n'.join(traceback.format_stack())
        if not is_py3k:
            exec('print >> sys.stderr, msg')
        else:
            exec('print(msg, file=sys.stderr)')
        run_setup(False)
    else:
        raise
